#  Copyright (c) 2021 zfit

import contextlib
import multiprocessing
import os
import sys
import warnings
from typing import List, Optional, Union

import tensorflow as tf
from dotmap import DotMap
from ordered_set import OrderedSet

from .deprecation import deprecated
from .exception import IllegalInGraphModeError
from .temporary import TemporarilySet


class RunManager:
    DEFAULT_MODE = {'graph': 'auto',
                    'autograd': True}

    def __init__(self, n_cpu='auto'):
        """Handle the resources and runtime specific options.

        The `run` method is equivalent to `sess.run`
        """
        self.MAX_CHUNK_SIZE = sys.maxsize
        self.chunking = DotMap()
        self._cpu = []
        self._n_cpu = None
        self._inter_cpus = None
        self._intra_cpus = None
        self._strict = False
        self.numeric_checks = True
        self._mode = self.DEFAULT_MODE.copy()
        self.set_n_cpu(n_cpu=n_cpu)

        # HACK
        self._enable_parameter_autoconversion = True
        # HACK END

        # set default values
        self.chunking.active = False  # not yet implemented the chunking...
        self.chunking.max_n_points = 1000000

    @property
    def mode(self):
        return self._mode

    @property
    def chunksize(self):
        if self.chunking.active:
            return self.chunking.max_n_points
        else:
            return self.MAX_CHUNK_SIZE

    @property
    def n_cpu(self):
        return len(self._cpu)

    def set_n_cpu(self, n_cpu: Union[str, int] = 'auto', strict: bool = False) -> None:
        """Set the number of cpus to be used by zfit. For more control, use `set_cpus_explicit`.

        Args:
            n_cpu: Number of cpus, will be the number for inter-op parallelism
            strict: If strict, sets intra parallelism to 1
        """
        if n_cpu == 'auto':
            try:
                cpu = sorted(os.sched_getaffinity(0))
            except AttributeError:
                cpu = range(multiprocessing.cpu_count())
                warnings.warn("Not running on Linux. Determining available cpus for thread can fail"
                              "and be overestimated. Workaround (only if too many cpus are used):"
                              "`zfit.run.set_n_cpu(your_cpu_number)`")
        elif isinstance(n_cpu, int):
            cpu = range(n_cpu)
        self._cpu = [f'dummy_cpu{i}' for i in cpu]
        n_cpu = len(cpu)
        if strict ^ self._strict:
            intra = 1 if strict else 2
            inter = n_cpu
            self.set_cpus_explicit(intra=intra, inter=inter)

    def set_cpus_explicit(self, intra: int, inter: int) -> None:
        """Set the number of threads (cpus) used for inter-op and intra-op parallelism.

        Args:
            intra: Number of threads used to perform an operation. For larger operations, e.g. large Tensors, this
                is usually beneficial to have >= 2.
            inter: Parallelization on the level of ops. This is beneficial, if many operations can be computed
                independently in parallel.
        """
        try:
            tf.config.threading.set_intra_op_parallelism_threads(intra)
            tf.config.threading.set_inter_op_parallelism_threads(inter)
            self._n_cpu = inter + intra
        except RuntimeError as err:
            raise RuntimeError("Cannot set the number of cpus after initialization, has to be at the beginning."
                               f" Original message: {err}")

    @contextlib.contextmanager
    def aquire_cpu(self, max_cpu: int = -1) -> List[str]:
        if isinstance(max_cpu, int):
            if max_cpu < 0:
                max_cpu = max((self.n_cpu + 1 + max_cpu, 0))  # -1 means all
            if max_cpu == 0:
                cpu = []
            else:
                n_cpu = min((max_cpu, self.n_cpu))

                cpu = self._cpu[-n_cpu:]
                self._cpu = self._cpu[:-n_cpu]

            yield cpu
            self._cpu.extend(cpu)

    def __call__(self, *args, **kwargs):
        # TODO: catch maybe sets, as they change the number of elements if we have identical ones
        # and convert them. Before it's fine, e.g. Parameters are unique, but after it's a value.
        if kwargs:
            raise RuntimeError("Why kwargs provided?")

        flattened_args = tf.nest.flatten(args)
        evaluated_args = [eval_object(arg) for arg in flattened_args]
        values = tf.nest.pack_sequence_as(args, flat_sequence=evaluated_args)

        # was_container = is_container(args[0]) and not isinstance(args[0], np.ndarray, )
        # if not was_container and values:
        #     values = values[0]
        if len(args) == 1:
            values = values[0]
        return values

    @staticmethod
    @deprecated(date=None, instructions="Use `set_graph_mode(False)`")
    def experimental_enable_eager(eager: bool = False):
        """DEPRECEATED! Enable eager makes tensorflow run like numpy. Useful for debugging.

        Do NOT directly mix it with Numpy (and if, also enable the numberical gradient).

        This can BREAK in the future.
        """

        from .graph import jit
        jit._set_all(not eager)

    def set_graph_mode(self, graph: Optional[Union[bool, str, dict]] = None):
        """Set the policy for graph building and the usage of automatic vs numerical gradients.

        zfit runs on top of TensorFlow, a modern, powerful computing engine very similar in design to Numpy.
        An interactive tutorial can be found at https://github.com/zfit/zfit-tutorials

        **Graph building**

        It has two ways to be run where the first
        defaults to the normal mode we are in except inside a :py:func:`~zfit.z.function` decorated function. Setting
        the mode allows to control the behavior of decorated functions to not always trigger a graph building.

         - **numpy-like/eager**: in this mode, the syntax slightly differs from pure numpy but is similar. For example,
            `tf.sqrt`, `tf.math.log` etc. The return values are `EagerTensors` that represent "wrapped Numpy arrays" and
            can directly be used with any Numpy function. They can explicitly be converted to a Numpy array with
            `zfit.run(EagerTensor)`, which takes also care of nested structures and already existing `np.ndarrays`,
            or just a `.numpy()` method.
            The difference to Numpy is that TensorFlow tries to optimize the calculation slightly beforehand and may
            also executes on the GPU. This will result in a slight performance penalty for *very small* computations
            compared to Numpy, on the other hand an improved performance for larger computations.

         - **graph**: a function can be decorated with :py:func:`~zfit.z.function`, which will *not* execute its
            content immediately, but first trace it and build a graph. This is done by *recording all `tf.*` operations
            and adding them to the graph while any Python operation, e.g. `np.random.*` will use a fixed value added
            to the graph. Building a graph greatly reduces the flexibility, since only `tf.*` operations can
            effectively be used to have dynamics in there, on the other hand it can greatly increase the performance.
            When the graph is built, it is cached (for later re-use), optimized and then executed.
            Calling a `tf.function` decorated
            function does therefore not make an actualy difference *for the caller*. But it is a difference on how the
            function behaves.

            .. code-block:: python

                @z.function
                def add_rnd(x):
                     res1 = x + np.random.uniform()  # returns a Python scalar. This exact scalar will be constant
                     res2 = x + z.random.uniform(shape=())  # returns a tf.Tensor. This will be flexible
                     return res1, res2

                res_np1, res_tf1 = add_rnd(5)
                res_np2, res_tf2 = add_rnd(5)

                assert res_np1 == res_np2  # they will be the same!
                assert res_tf1 != res_tf2  # these differ

            While writing TensorFlow is just like Numpy, if we build a graph, only `tf.*` dynamics "survives".
            Important: while values are usually constant, changing a :py:class:`zfit.Parameter` value with
            :py:meth:`~zfit.Parameter.set_value(...)` *will* change the value in the graph as well.

                        .. code-block:: python

                @z.function
                def add(x, param):
                     return x + param

                param = zfit.Parameter('param1', 36)
                assert add_rnd(5, param) == 41
                param.set_value(6)
                assert add_rnd(5, param) == 42  # the value changed!

            Every graph generation takes some additional time and is stored, consuming memory and slowing down the
            overall execution process.
            To clear all caches and force a rebuild of the graph, `zfit.run.clear_graph_cache()` can be used.

            If a function is not decorated with `z.function`, this does not guarantee that it is executed in eager,
            as an outer function may uses a decorator. A typical case is the loss, which is decorated. Therefore,
            any Model called inside will be evaluated with a graph building first.


            **When to use what**:
             - Any repeated call (as a typical call to the loss function in the minimization process) is usually
               better suited within a `z.function`.
             - A single call (e.g. for plotting) or repeated calls *with different arguments* should rather be run
               *without* a graph built first
             - Debugging is usually way easier without graph building. Therefore, set the graph mode to `False`
             - If the minimization fails but the pdf works without graph, maybe the graph mode can be switched on
               for everything to have the same behavior in the pdf as when the loss is called.

        Args:
            graph: Policy for when to build a graph with which function. Currently allowed values are
              - `True`: this will make all :py:func:`zfit.z.function` decorated function to be traced. Useful
                to have a consistent behavior overall, as e.g. a PDF may not be traced if `pdf` or `integrate` is
                called, but may be traced when inside a loss.
              - `False`: this will make everything execute immediately, like Numpy (this is **not enough** to be
                fully Numpy compatible in the sense of using `, also see the `autograd` option)
              - 'auto': Something in between, where sampling (currently) and the loss builds a graph but
                all model methods, such as `pdf`, `integrate` (except of `*sample*`) do not and are executed
                eagerly.
              - (**advanced and experimental!**): a dictionary containing the string of a wrapped function identifier
                (see also :py:func:`~zfit.z.function` for more information about this) with a boolean that switches
                explicitly on/off the graph building for this type of decorated functions.
        """
        if not tf.executing_eagerly():
            raise IllegalInGraphModeError("Cannot change the execution mode of graph inside a `z.function`"
                                          " decorated function. Only possible in an eager context.")
        if graph is None:
            graph = 'auto'
        return TemporarilySet(value=graph, setter=self._set_graph_mode, getter=self.get_graph_mode)

    def set_autograd_mode(self, autograd: Optional[bool] = None):
        """Use automatic or numerical gradients.

        zfit runs on top of TensorFlow, a modern, powerful computing engine very similar in design to Numpy.
        An interactive tutorial can be found at https://github.com/zfit/zfit-tutorials

        **automatic gradient**

        A strong feature of TensorFlow is the possibility to derive an analytic expression for the gradient
        by successively applying the chain rule to all of its operations. This is *independent* of whether the code
        is run in graph or eager execution, but requires all operations that are dynamic to be `tf.*` operations.
        For example, multiplying by a constant (constant as in *not chaning ever*) does not require the constant to
        be a `tf.constant(...)` but can be a Python scalar. For example, it is also fine to use a fixed template shape
        using Numpy (Scipy etc), as the template shape will stay constant (this requires though to use a
        `z.numpy_function` to work, but this is another story about graph mode or not).

        To allow to have dynamic numpy operations in a component, preferably wrapped with `z.numpy_function` instead of
        forced eager, and to still retrieve a meaningful gradient, a numerical gradient has to be used.
        In general, this can be achieved by setting the `autograd` to False. Any derivative received will then be
        numerically computed. Furthermore, some minimizers (e.g. :py:class:`~zfit.minimize.Minuit`) have their own way
        of calculating gradients, which can be faster.
        Disabling `autograd` and using the zfit builting numerical way of calculating the gradients and hessian can
        be less stable and may raises errors.

        Args:
            autograd: Whether the automatic gradient feature of TensorFlow should be used or a numerical procedure
              instead. If any non-constant Python (numpy, scipy,...) code is used inside, this should be switched on.
        """
        if autograd is None:
            autograd = True
        return TemporarilySet(value=autograd, setter=self._set_autograd_mode, getter=self.get_autograd_mode)

    @deprecated(None, "Use `set_graph_mode` or `set_autograd_mode`.")
    def set_mode(self, graph: Optional[Union[bool, str, dict]] = None, autograd: Optional[bool] = None):
        """DEPRECATED!

        Use `set_graph_mode` or `set_autograd_mode`.
        """
        if autograd is not None:
            self._set_autograd_mode(autograd)
        if graph is not None:
            self._set_graph_mode(graph)

    def _set_autograd_mode(self, autograd):
        if autograd is not None:
            from zfit import settings
            settings.options.numerical_grad = not autograd
            self._mode['autograd'] = autograd

    def _set_graph_mode(self, graph):
        from .graph import jit as jit_obj

        # only run eagerly if no graph
        tf.config.run_functions_eagerly(graph is False)
        if graph is True:
            jit_obj._set_all(True)
        elif graph is False:
            jit_obj._set_all(False)
        elif graph == 'auto':
            jit_obj._set_default()
        elif isinstance(graph, dict):
            jit_obj._update_allowed(graph)
        elif graph is not None:
            raise ValueError(f"{graph} is not a valid keyword to the `jit` behavior. Use either "
                             f"True, False, 'default' or a dict. You can read more about it in the docs.")
        if graph is not None:
            self._mode['graph'] = graph

    def get_graph_mode(self) -> Union[bool, str]:
        """Return the current policy for graph building.

        Returns:
            The current policy. For more information, check :py:meth:`~zfit.run.set_mode`.
        """
        return self.mode['graph']

    @deprecated(None, 'Use `get_graph_mode` instead.')
    def current_policy_graph(self) -> Union[bool, str]:
        """DEPRECEATED!

        Use `get_graph_mode` instead.
        """
        return self.get_graph_mode()

    def get_autograd_mode(self) -> bool:
        """The current policy for using the automatic gradient or falling back to the numerical.

        Returns:
            If autograd is being used.
        """
        return self.mode['autograd']

    def current_policy_autograd(self) -> bool:
        """DEPRECATED!

        Use `get_autograd_mode` instead.
        """
        return self.get_autograd_mode()

    def set_mode_default(self):
        """Reset the mode to the default of `graph` = 'auto' and `autograd` = True."""

        return TemporarilySet(value=self.DEFAULT_MODE,
                              setter=lambda v: self.set_mode(**v),
                              getter=lambda: self._mode)

    def clear_graph_cache(self):
        """Clear all generated graphs and effectively reset. Should not affect execution, only performance.

        In a simple fit scenario, this is not used. But if several fits are performed with different python objects
        such as a scan over a range (by changing the norm_range and creating a new dataset), doing minimization and
        therefore invoking the loss (by default creating a graph) will leave the graphs in the cache, even tough
        the already scanned ranges are not needed anymore.

        To clean, this function can be invoked. The only effect should be to speed up things, but should not have
        any side-effects other than that.
        """
        from zfit.util.cache import clear_graph_cache
        clear_graph_cache()

    def assert_executing_eagerly(self):
        """Assert that the execution is eager and Python side effects are taken into account.

        This can be placed inside a model _in case python side-effects are necessary_ and no other way is possible.
        """
        if not tf.executing_eagerly():
            raise RuntimeError("This code is not supposed to run inside a graph.")

    @property
    @deprecated(None, "Use `current_policy_graph() is False`")
    def experimental_is_eager(self):
        return tf.executing_eagerly()

    def executing_eagerly(self):
        return tf.executing_eagerly()

    @deprecated(date=None, instructions="Use clear_graph_caches instead.")
    def experimental_clear_caches(self):
        """DEPRECATED!

        Use `clear_graph_caches` instead.
        """
        self.clear_graph_cache()


def eval_object(obj: object) -> object:
    from zfit.core.parameter import BaseComposedParameter
    if isinstance(obj, BaseComposedParameter):  # currently no numpy attribute. Should we add this?
        obj = obj.value()
    if tf.is_tensor(obj):
        return obj.numpy()
    else:
        return obj



#  Copyright (c) 2021 zfit
""".. todo:: Add module docstring."""
from .util.checks import NONE
from .util.deprecation import deprecated
from .util.exception import (AnalyticIntegralNotImplemented,
                             AnalyticNotImplemented,
                             AnalyticSamplingNotImplemented,
                             FunctionNotImplemented, IllegalInGraphModeError,
                             InitNotImplemented, MaximumIterationReached,
                             MinimizerSubclassingError,
                             MultipleLimitsNotImplemented,
                             NameAlreadyTakenError, NormRangeNotImplemented,
                             SpecificFunctionNotImplemented,
                             VectorizedLimitsNotImplemented)

__all__ = ['NameAlreadyTakenError', 'IllegalInGraphModeError', 'NormRangeNotImplemented',
           'MultipleLimitsNotImplemented', 'VectorizedLimitsNotImplemented',
           'AnalyticNotImplemented', 'FunctionNotImplemented', 'AnalyticSamplingNotImplemented',
           'AnalyticIntegralNotImplemented', 'SpecificFunctionNotImplemented', 'MinimizerSubclassingError',
           "MaximumIterationReached",
           'InitNotImplemented', 'NONE']


class VectorizedLimitsNotImplementedError(VectorizedLimitsNotImplemented):
    @deprecated(None, 'Use VectorizedLimitsNotImplemented instead.')
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class SpecificFunctionNotImplementedError(SpecificFunctionNotImplemented):
    @deprecated(None, 'Use SpecificFunctionNotImplemented instead.')
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class NormRangeNotImplementedError(NormRangeNotImplemented):
    @deprecated(None, 'Use NormRangeNotImplemented instead.')
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class MultipleLimitsNotImplementedError(MultipleLimitsNotImplemented):
    @deprecated(None, 'Use MultipleLimitsNotImplemented instead.')
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class FunctionNotImplementedError(FunctionNotImplemented):
    @deprecated(None, 'Use FunctionNotImplemented instead.')
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticSamplingNotImplementedError(AnalyticSamplingNotImplemented):
    @deprecated(None, 'Use AnalyticSamplingNotImplemented instead.')
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticIntegralNotImplementedError(AnalyticIntegralNotImplemented):
    @deprecated(None, 'Use AnalyticIntegralNotImplemented instead.')
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticNotImplementedError(AnalyticNotImplemented):
    @deprecated(None, 'Use AnalyticNotImplemented instead.')
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



#  Copyright (c) 2021 zfit

from .core.basepdf import BasePDF
from .models.basic import Exponential
from .models.conditional import ConditionalPDFV1
from .models.convolution import FFTConvPDFV1
from .models.dist_tfp import (Cauchy, Gauss, Poisson, TruncatedGauss, Uniform,
                              WrapDistribution)
from .models.functor import BaseFunctor, ProductPDF, SumPDF
from .models.kde import GaussianKDE1DimV1
from .models.physics import CrystalBall, DoubleCB
from .models.polynomials import (Chebyshev, Chebyshev2, Hermite, Laguerre,
                                 Legendre, RecursivePolynomial)
from .models.special import ZPDF, SimpleFunctorPDF, SimplePDF

__all__ = ['BasePDF', 'BaseFunctor',
           'Exponential',
           'CrystalBall', 'DoubleCB',
           'Gauss', 'Uniform', 'TruncatedGauss', 'WrapDistribution', 'Cauchy', 'Poisson',
           "Chebyshev", "Legendre", "Chebyshev2", "Hermite", "Laguerre", "RecursivePolynomial",
           'ProductPDF', 'SumPDF',
           'GaussianKDE1DimV1',
           'FFTConvPDFV1',
           'ConditionalPDFV1',
           'ZPDF', 'SimplePDF', 'SimpleFunctorPDF'
           ]



#  Copyright (c) 2021 zfit

import abc
from abc import abstractmethod


class ZfitResult:
    @abstractmethod
    def hesse(self, params, method):
        """Calculate for `params` the symmetric error using the Hessian matrix.

        Args:
            params: The parameters  to calculate the
                Hessian symmetric error. If None, use all parameters.
            method: the method to calculate the hessian. Can be {'minuit'} or a callable.

        Returns:
            Result of the hessian (symmetric) error as dict with each parameter holding
                the error dict {'error': sym_error}.

                So given param_a (from zfit.Parameter(.))
                `error_a = result.hesse(params=param_a)[param_a]['error']`
                error_a is the hessian error.
        """
        raise NotImplementedError

    @abstractmethod
    def errors(self, params, method, cl):
        """Calculate and set for `params` the asymmetric error using the set error method.

        Args:
            params: The parameters or their names to calculate the
                 errors. If `params` is `None`, use all *floating* parameters.
            method: The method to use to calculate the errors. Valid choices are
                {'minuit_minos'} or a Callable.

        Returns:
            A `OrderedDict` containing as keys the parameter names and as value a `dict` which
                contains (next to probably more things) two keys 'lower' and 'upper',
                holding the calculated errors.
                Example: result['par1']['upper'] -> the asymmetric upper error of 'par1'
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def minimizer(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def params(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def fmin(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def loss(self):
        raise NotImplementedError


class ZfitMinimizer(abc.ABC):
    """Define the minimizer interface."""

    @abc.abstractmethod
    def minimize(self, loss, params=None, init=None):
        raise NotImplementedError

    def step(self, loss, params=None):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def tol(self):
        raise NotImplementedError



#  Copyright (c) 2021 zfit
import platform

import numpy as np
import pytest

import zfit
from zfit import z
from zfit.minimizers.errors import compute_errors
from zfit.minimizers.fitresult import FitResult

true_a = 3.
true_b = 1.1
true_c = -0.3

true_val = [true_a, true_b, true_c]


def create_loss(n=15000, weights=None):
    avalue = 1.5
    a_param = zfit.Parameter("variable_a15151", avalue, -1., 20.,
                             step_size=z.constant(0.1))
    a_param.init_val = avalue
    bvalue = 1.5
    b_param = zfit.Parameter("variable_b15151", bvalue, 0, 20)
    b_param.init_val = bvalue
    cvalue = -0.04
    c_param = zfit.Parameter("variable_c15151", cvalue, -1, 0.)
    c_param.init_val = cvalue
    obs1 = zfit.Space(obs='obs1', limits=(-2.4, 9.1))

    # load params for sampling
    a_param.set_value(true_a)
    b_param.set_value(true_b)
    c_param.set_value(true_c)

    gauss1 = zfit.pdf.Gauss(mu=a_param, sigma=b_param, obs=obs1)
    exp1 = zfit.pdf.Exponential(lam=c_param, obs=obs1)

    sum_pdf1 = zfit.pdf.SumPDF((gauss1, exp1), 0.2)

    sampled_data = sum_pdf1.create_sampler(n=n)
    sampled_data.resample()

    if weights is not None:
        sampled_data.set_weights(weights)

    loss = zfit.loss.UnbinnedNLL(model=sum_pdf1, data=sampled_data)

    return loss, (a_param, b_param, c_param)


def create_fitresult(minimizer_class_and_kwargs, n=15000, weights=None):
    loss, (a_param, b_param, c_param) = create_loss(n=n, weights=weights)

    true_minimum = loss.value().numpy()

    for param in [a_param, b_param, c_param]:
        param.assign(param.init_val)  # reset the value

    minimizer_class, minimizer_kwargs, test_error = minimizer_class_and_kwargs
    minimizer = minimizer_class(**minimizer_kwargs)

    result = minimizer.minimize(loss=loss)
    cur_val = loss.value().numpy()
    aval, bval, cval = (v.numpy() for v in (a_param, b_param, c_param))

    ret = {'result': result, 'true_min': true_minimum, 'cur_val': cur_val, 'a': aval, 'b': bval, 'c': cval,
           'a_param': a_param, 'b_param': b_param, 'c_param': c_param}

    return ret


def test_set_values():
    upper1 = 5.33
    lower1 = 0.
    param1 = zfit.Parameter('param1', 2., lower1, upper1)
    param1.set_value(lower1)
    assert pytest.approx(zfit.run(param1.value()), lower1)
    param1.set_value(upper1)
    assert pytest.approx(zfit.run(param1.value()), upper1)
    with pytest.raises(ValueError):
        param1.set_value(lower1 - 0.001)
    with pytest.raises(ValueError):
        param1.set_value(upper1 + 0.001)

    fitresult = create_fitresult((zfit.minimize.Minuit, {}, True))
    result = fitresult['result']
    param_a = fitresult['a_param']
    param_b = fitresult['b_param']
    param_c = fitresult['c_param']

    val_a = fitresult['a']
    val_b = fitresult['b']
    val_c = fitresult['c']
    with pytest.raises(ValueError):
        param_b.set_value(999)
    param_c.assign(9999)
    zfit.param.set_values([param_c, param_b], values=result)

    assert param_a.value() == val_a
    assert param_b.value() == val_b
    assert param_c.value() == val_c

    param_d = zfit.Parameter("param_d", 12)
    with pytest.raises(ValueError):
        zfit.param.set_values([param_d], result)


minimizers = [
    (zfit.minimize.NLoptLBFGSV1, {}, True),
    (zfit.minimize.ScipyTrustConstrV1, {}, True),
    (zfit.minimize.Minuit, {}, True),
]
if not platform.system() == 'Darwin':  # TODO: Ipyopt installation on macosx not working
    minimizers.append((zfit.minimize.IpyoptV1, {}, False))
# sort for xdist: https://github.com/pytest-dev/pytest-xdist/issues/432
minimizers = sorted(minimizers, key=lambda val: repr(val))


@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers)
def test_fmin(minimizer_class_and_kwargs):
    results = create_fitresult(minimizer_class_and_kwargs=minimizer_class_and_kwargs)
    result = results['result']
    assert pytest.approx(results['cur_val']) == result.fmin


@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers)
def test_params(minimizer_class_and_kwargs):
    results = create_fitresult(minimizer_class_and_kwargs=minimizer_class_and_kwargs)
    result = results['result']
    np.testing.assert_allclose(true_val, result.values, rtol=0.2)
    for param in result.params:
        assert result.values[param] == result.params[param]['value']
        assert result.values[param.name] == result.params[param]['value']
        assert result.values[param.name] == result.params[param.name]['value']


def test_params_at_limit():
    loss, (param_a, param_b, param_c) = create_loss(n=5000)
    old_lower = param_a.lower
    param_a.lower = param_a.upper
    param_a.assign(param_a.upper + 5)
    minimizer = zfit.minimize.Minuit(gradient=True, tol=10.)
    result = minimizer.minimize(loss)
    param_a.assign(100)
    assert param_a.at_limit
    assert result.params_at_limit
    param_a.lower = old_lower
    assert param_a.at_limit
    assert not result.valid


@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers)
@pytest.mark.parametrize("use_weights", [False, True])
def test_covariance(minimizer_class_and_kwargs, use_weights):
    n = 15000
    if use_weights:
        weights = np.random.normal(1, 0.001, n)
    else:
        weights = None

    results = create_fitresult(minimizer_class_and_kwargs=minimizer_class_and_kwargs,
                               n=n,
                               weights=weights)
    result = results['result']
    hesse = result.hesse()
    a = results['a_param']
    b = results['b_param']
    c = results['c_param']

    with pytest.raises(KeyError):
        result.covariance(params=[a, b, c], method="hesse")

    cov_mat_3 = result.covariance(params=[a, b, c])
    cov_mat_2 = result.covariance(params=[c, b])
    cov_dict = result.covariance(params=[a, b, c], as_dict=True)

    assert pytest.approx(hesse[a]['error'], rel=0.01) == np.sqrt(cov_dict[(a, a)])
    assert pytest.approx(hesse[a]['error'], rel=0.01) == np.sqrt(cov_mat_3[0, 0])

    assert pytest.approx(hesse[b]['error'], rel=0.01) == np.sqrt(cov_dict[(b, b)])
    assert pytest.approx(hesse[b]['error'], rel=0.01) == np.sqrt(cov_mat_3[1, 1])
    assert pytest.approx(hesse[b]['error'], rel=0.01) == np.sqrt(cov_mat_2[1, 1])

    assert pytest.approx(hesse[c]['error'], rel=0.01) == np.sqrt(cov_dict[(c, c)])
    assert pytest.approx(hesse[c]['error'], rel=0.01) == np.sqrt(cov_mat_3[2, 2])
    assert pytest.approx(hesse[c]['error'], rel=0.01) == np.sqrt(cov_mat_2[0, 0])

    if use_weights:
        rtol, atol = 0.15, 0.015
    else:
        rtol, atol = 0.05, 0.005

    cov_mat_3_np = result.covariance(params=[a, b, c], method="hesse_np")
    np.testing.assert_allclose(cov_mat_3, cov_mat_3_np, rtol=rtol, atol=atol)


@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers)
def test_correlation(minimizer_class_and_kwargs):
    results = create_fitresult(minimizer_class_and_kwargs=minimizer_class_and_kwargs)
    result = results['result']
    hesse = result.hesse()
    a = results['a_param']
    b = results['b_param']
    c = results['c_param']

    cor_mat = result.correlation(params=[a, b, c])
    cov_mat = result.covariance(params=[a, b, c])
    cor_dict = result.correlation(params=[a, b], as_dict=True)

    np.testing.assert_allclose(np.diag(cor_mat), 1.0)

    a_error = hesse[a]['error']
    b_error = hesse[b]['error']
    assert pytest.approx(cor_mat[0, 1], rel=0.01) == cov_mat[0, 1] / (a_error * b_error)
    assert pytest.approx(cor_dict[(a, b)], rel=0.01) == cov_mat[0, 1] / (a_error * b_error)


@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers)
@pytest.mark.parametrize("cl", [None, 0.683, 0.8, 0.95, 0.9])
@pytest.mark.timeout(60)  # if stuck finding new minima
def test_errors(minimizer_class_and_kwargs, cl):
    n_max_trials = 5  # how often to try to find a new minimum
    results = create_fitresult(minimizer_class_and_kwargs=minimizer_class_and_kwargs)
    result = results['result']
    a = results['a_param']
    b = results['b_param']
    c = results['c_param']

    for n_trial in range(n_max_trials):
        z_errors, new_result = result.errors(method="zfit_error", cl=cl)
        minos_errors, _ = result.errors(method="minuit_minos", cl=cl)
        if new_result is None:
            break
        else:
            result = new_result
    else:  # no break occured
        assert False, "Always a new minimum was found, cannot perform test."

    for param in [a, b, c]:
        z_error_param = z_errors[param]
        minos_errors_param = minos_errors[param]
        for dir in ["lower", "upper"]:
            assert pytest.approx(z_error_param[dir], rel=0.03) == minos_errors_param[dir]

    with pytest.raises(KeyError):
        result.errors(method="error")


@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers)
def test_new_minimum(minimizer_class_and_kwargs):
    loss, params = create_loss(10000)

    minimizer_class, minimizer_kwargs, test_error = minimizer_class_and_kwargs
    minimizer = minimizer_class(**minimizer_kwargs)

    a_param, b_param, c_param = params

    if test_error:

        b_param.floating = False
        b_param.set_value(3.7)
        c_param.floating = False
        result = minimizer.minimize(loss=loss)
        b_param.floating = True
        c_param.floating = True

        params_dict = {p: p.numpy() for p in params}
        hacked_result = FitResult(params=params_dict, edm=result.edm, fmin=result.fmin, info=result.info,
                                  loss=loss, status=result.status, converged=result.converged, valid=True,
                                  message="hacked for unittest", niter=999, criterion=None,
                                  minimizer=minimizer.copy())

        method = lambda **kwgs: compute_errors(covariance_method="hesse_np", **kwgs)

        errors, new_result = hacked_result.errors(params=params, method=method, name="interval")

        assert new_result is not None

        assert hacked_result.valid is False
        for p in params:
            assert errors[p] == "Invalid, a new minimum was found."

        assert new_result.valid is True
        errors, _ = new_result.errors()
        for param in params:
            assert errors[param]['lower'] < 0
            assert errors[param]['upper'] > 0



#  Copyright (c) 2021 zfit
import numpy as np
import pytest

import zfit
from zfit import Parameter
from zfit.models.dist_tfp import Poisson

lamb_true = 50

obs = zfit.Space(obs="Nobs", limits=(0, 200))

test_values = np.random.uniform(low=0, high=100, size=100)


def create_poisson():
    N = Parameter("N", lamb_true)
    poisson = Poisson(obs=obs, lamb=N)
    return poisson


def create_poisson_composed_rate():
    N1 = Parameter("N1", lamb_true / 2)
    N2 = Parameter("N2", lamb_true / 2)
    N = zfit.param.ComposedParameter("N", lambda n1, n2: n1 + n2, params=[N1, N2])

    poisson = Poisson(obs=obs, lamb=N)
    return poisson


@pytest.mark.parametrize('composed_rate', [False, True])
def test_poisson(composed_rate):
    if composed_rate:
        poisson = create_poisson_composed_rate()
    else:
        poisson = create_poisson()

    probs1 = poisson.pdf(x=test_values)
    probs1 = probs1.numpy()

    samples = poisson.sample(10000).numpy()

    assert np.std(samples) == pytest.approx(50 ** 0.5, rel=0.05)



#  Copyright (c) 2020 zfit
import abc
import itertools
from typing import Iterable

from ordered_set import OrderedSet

from zfit.core.interfaces import ZfitDependenciesMixin, ZfitObject
from zfit.util import ztyping
from zfit.util.container import convert_to_container


class BaseDependentsMixin(ZfitDependenciesMixin):
    @abc.abstractmethod
    def _get_dependencies(self) -> ztyping.DependentsType:
        raise NotImplementedError

    def get_cache_deps(self, only_floating: bool = True) -> ztyping.DependentsType:
        """Return a set of all independent :py:class:`~zfit.Parameter` that this object depends on.

        Args:
            only_floating: If `True`, only return floating :py:class:`~zfit.Parameter`
        """
        dependencies = self._get_dependencies()
        if only_floating:
            dependencies = OrderedSet(filter(lambda p: p.floating, dependencies))
        return dependencies


def _extract_dependencies(zfit_objects: Iterable[ZfitObject]) -> ztyping.DependentsType:
    """Calls the :py:meth:`~BaseDependentsMixin.get_dependents` method on every object and returns a combined set.

    Args:
        zfit_objects:

    Returns:
        A set of independent Parameters
    """
    zfit_objects = convert_to_container(zfit_objects)
    dependents = (obj.get_cache_deps(only_floating=False) for obj in zfit_objects)
    dependents_set = OrderedSet(itertools.chain.from_iterable(dependents))  # flatten
    return dependents_set



#  Copyright (c) 2021 zfit

from typing import Callable, Dict, Iterable, Union

import tensorflow as tf

import zfit.z.numpy as znp

from ..core.basefunc import BaseFunc
from ..core.basemodel import SimpleModelSubclassMixin
from ..core.dependents import _extract_dependencies
from ..core.interfaces import ZfitFunc, ZfitModel
from ..core.space import supports
from ..models.basefunctor import FunctorMixin
from ..util import ztyping
from ..util.container import convert_to_container


class SimpleFunc(BaseFunc):

    def __init__(self, obs: ztyping.ObsTypeInput, func: Callable, name: str = "Function", **params):
        """Create a simple function out of of `func` with the observables `obs` depending on `parameters`.

        Args:
            func:
            obs:
            name:
            **params: The parameters as keyword arguments. E.g. `mu=Parameter(...)`
        """
        super().__init__(name=name, obs=obs, params=params)
        self._value_func = self._check_input_x_function(func)

    def _func(self, x):
        try:
            return self._value_func(x)
        except TypeError:  # self requested, TODO maybe check signature?
            return self._value_func(self, x)


class BaseFunctorFunc(FunctorMixin, BaseFunc):
    def __init__(self, funcs, name="BaseFunctorFunc", params=None, **kwargs):
        funcs = convert_to_container(funcs)
        if params is None:
            params = {}
        # for func in funcs:
        #     params.update(func.params)

        self.funcs = funcs
        super().__init__(name=name, models=self.funcs, params=params, **kwargs)

    def _get_dependencies(self):  # TODO: change recursive to `only_floating`?
        dependents = super()._get_dependencies()  # get the own parameter dependents
        func_dependents = _extract_dependencies(self.funcs)  # flatten
        return dependents.union(func_dependents)

    @property
    def _models(self) -> Dict[Union[float, int, str], ZfitModel]:
        return self.funcs


class SumFunc(BaseFunctorFunc):
    def __init__(self, funcs: Iterable[ZfitFunc], obs: ztyping.ObsTypeInput = None, name: str = "SumFunc", **kwargs):
        super().__init__(funcs=funcs, obs=obs, name=name, **kwargs)

    def _func(self, x):
        # sum_funcs = tf.add_n([func.value(x) for func in self.funcs])
        funcs = [func.func(x) for func in self.funcs]
        sum_funcs = tf.math.accumulate_n(funcs)
        return sum_funcs

    @supports()
    def _analytic_integrate(self, limits, norm_range):
        # below may raises AnalyticIntegralNotImplementedError, that's fine. We don't wanna catch that.
        integrals = [func.analytic_integrate(limits=limits, norm_range=norm_range) for func in self.funcs]
        return tf.math.accumulate_n(integrals)


class ProdFunc(BaseFunctorFunc):
    def __init__(self, funcs: Iterable[ZfitFunc], obs: ztyping.ObsTypeInput = None, name: str = "SumFunc", **kwargs):
        super().__init__(funcs=funcs, obs=obs, name=name, **kwargs)

    def _func(self, x):
        funcs = [func.func(x) for func in self.funcs]
        product = znp.prod(funcs, axis=0)
        return product


class ZFunc(SimpleModelSubclassMixin, BaseFunc):
    def __init__(self, obs: ztyping.ObsTypeInput, name: str = "ZFunc", **params):
        super().__init__(obs=obs, name=name, **params)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._check_simple_model_subclass()



#  Copyright (c) 2020 zfit

import numpy as np

import zfit

# create space
xobs = zfit.Space('xobs', (-4, 4))
yobs = zfit.Space('yobs', (-3, 5))
zobs = zfit.Space('z', (-2, 4))
obs = xobs * yobs * zobs

# parameters
mu1 = zfit.Parameter("mu1", 1., -4, 6)
mu23 = zfit.Parameter("mu_shared", 1., -4, 6)
sigma12 = zfit.Parameter("sigma_shared", 1., 0.1, 10)
sigma3 = zfit.Parameter("sigma3", 1., 0.1, 10)

# model building, pdf creation
gauss_x = zfit.pdf.Gauss(mu=mu1, sigma=sigma12, obs=xobs)
gauss_y = zfit.pdf.Gauss(mu=mu23, sigma=sigma12, obs=yobs)
gauss_z = zfit.pdf.Gauss(mu=mu23, sigma=sigma3, obs=zobs)

product_gauss = zfit.pdf.ProductPDF([gauss_x, gauss_y, gauss_z])

# OR create directly from your 3 dimensional pdf as
# model = MyPDF(obs=obs, param1=..., param2,...)

# data
normal_np = np.random.normal(loc=[2., 2.5, 2.5], scale=[3., 3, 1.5], size=(10000, 3))
data_raw = zfit.Data.from_numpy(obs=obs, array=normal_np)  # or from anywhere else, e.g. root

df = data_raw.to_pandas()
# preprocessing here, rename things. Match column names with the observable names "xobs", "yobs", "z" (they have to be
# contained, more columns in the df is not a problem)
data = zfit.Data.from_pandas(df, obs=obs)

# create NLL
nll = zfit.loss.UnbinnedNLL(model=product_gauss, data=data)

# create a minimizer
minimizer = zfit.minimize.Minuit()
result = minimizer.minimize(nll)
print(result.params)

# do the error calculations, here with minos
param_errors, _ = result.errors()
print(param_errors)



#  Copyright (c) 2021 zfit

import tensorflow as tf

import zfit
from zfit import z


class CustomPDF2D(zfit.pdf.BasePDF):
    """My custom, 2 dimensional pdf where the axes are: Energy, Momentum."""

    def __init__(self, param1, param2, param3, obs, name="CustomPDF", ):
        # we can now do complicated stuff here if needed
        # only thing: we have to specify explicitly here what is which parameter
        params = {'super_param': param1,  # we can change/compose etc parameters
                  'param2': param2, 'param3': param3}
        super().__init__(obs, params, name=name)

    def _unnormalized_pdf(self, x):
        energy, momentum = x.unstack_x()
        param1 = self.params['super_param']
        param2 = self.params['param2']
        param3 = self.params['param3']

        # just a fantasy function
        probs = param1 * tf.cos(energy ** 2) + tf.math.log(param2 * momentum ** 2) + param3
        return probs


# add an analytic integral

# define the integral function
def integral_full(limits, norm_range, params, model):
    lower, upper = limits.rect_limits    # for a more detailed guide, see the space.py example
    param1 = params['super_param']
    param2 = params['param2']
    param3 = params['param3']

    lower = z.convert_to_tensor(lower)
    upper = z.convert_to_tensor(upper)

    # calculate the integral here, dummy integral, wrong!
    integral = param1 * param2 * param3 + z.reduce_sum([lower, upper])
    return integral


# define the space over which it is defined. Here, we use the axes
lower_full = (-10, zfit.Space.ANY_LOWER)
upper_full = (10, zfit.Space.ANY_UPPER)
integral_full_limits = zfit.Space(axes=(0, 1),
                                  limits=(lower_full, upper_full))

CustomPDF2D.register_analytic_integral(func=integral_full,
                                       limits=integral_full_limits)


# define the partial integral function
def integral_axis1(x, limits, norm_range, params, model):
    data_0 = x.unstack_x()  # data from axis 0

    param1 = params['super_param']
    param2 = params['param2']
    param3 = params['param3']

    lower, upper = limits.limit1d  # for a more detailed guide, see the space.py example
    lower = z.convert_to_tensor(lower)  # the limits are now 1-D, for axis 1
    upper = z.convert_to_tensor(upper)

    # calculate the integral here, dummy integral
    integral = data_0 ** 2 * param1 * param2 * param3 + z.reduce_sum([lower, upper])
    # notice that the returned shape will be in the same as data_0, e.g. the number of events given in x
    return integral


# define the space over which it is defined. Here, we use the axes
lower_axis1 = ((zfit.Space.ANY_LOWER,),)
upper_axis1 = ((zfit.Space.ANY_UPPER,),)
integral_axis1_limits = zfit.Space(axes=(1,),  # axes one corresponds to the second obs, here obs2
                                   limits=(lower_axis1, upper_axis1))

CustomPDF2D.register_analytic_integral(func=integral_axis1,
                                       limits=integral_axis1_limits)

if __name__ == '__main__':
    import numpy as np

    obs = zfit.Space('obs1', (-10, 10)) * zfit.Space('obs2', (-3, 5))
    pdf = CustomPDF2D(1, 2, 3, obs=obs)
    sample = pdf.sample(n=1000)
    pdf.pdf([[2., 2.5], [5.4, 3.2]])
    x_part = zfit.Data.from_numpy(array=np.array([2.1, 2.2, 3.2]), obs='obs1')

    # integrate over obs2 with limits 1, 2 for the `x_part`. This will use the analytic integral above
    pdf.partial_integrate(x=x_part, limits=zfit.Space('obs2', (1, 2)))
    # we can explicitly call the analytic integral. Without registering it (e.g. comment the line with the `register`
    # and run again), it will raise an error
    pdf.partial_analytic_integrate(x=x_part, limits=zfit.Space('obs2', (1, 2)))



#  Copyright (c) 2021 zfit
from typing import List

import numpy as np
import pytest
import tensorflow as tf

import zfit
from zfit.core.data import Data
from zfit.core.interfaces import ZfitPDF
from zfit.core.parameter import Parameter
from zfit.core.space import Space
from zfit.models.dist_tfp import Gauss
from zfit.models.functor import ProductPDF, SumPDF

low, high = -0.64, 5.9

mu1_true = 1.
mu2_true = 2.
mu3_true = 0.6
sigma1_true = 1.4
sigma2_true = 2.3
sigma3_true = 1.8

fracs = [0.3, 0.15]

obs1 = zfit.Space('obs1', (-3, 6))
obs2 = zfit.Space('obs2', (-2, 5))
obs3 = zfit.Space('obs3', (-3.5, 4))


def true_gaussian_sum(x):
    def norm(sigma):
        return np.sqrt(2 * np.pi) * sigma

    sum_gauss = fracs[0] * np.exp(- (x - mu1_true) ** 2 / (2 * sigma1_true ** 2)) / norm(sigma1_true)
    sum_gauss += fracs[1] * np.exp(- (x - mu2_true) ** 2 / (2 * sigma2_true ** 2)) / norm(sigma2_true)
    sum_gauss += (1. - sum(fracs)) * np.exp(- (x - mu3_true) ** 2 / (2 * sigma3_true ** 2)) / norm(sigma3_true)
    return sum_gauss


def create_params(name_add=""):
    mu1 = Parameter("mu1" + name_add, mu1_true)
    mu2 = Parameter("mu2" + name_add, mu2_true)
    mu3 = Parameter("mu3" + name_add, mu3_true)
    sigma1 = Parameter("sigma1" + name_add, sigma1_true)
    sigma2 = Parameter("sigma2" + name_add, sigma2_true)
    sigma3 = Parameter("sigma3" + name_add, sigma3_true)
    return mu1, mu2, mu3, sigma1, sigma2, sigma3


def create_gaussians() -> List[ZfitPDF]:
    # Gauss for sum, same axes
    mu1, mu2, mu3, sigma1, sigma2, sigma3 = create_params()
    gauss1 = Gauss(mu=mu1, sigma=sigma1, obs=obs1, name="gauss1asum")
    gauss2 = Gauss(mu=mu2, sigma=sigma2, obs=obs1, name="gauss2asum")
    gauss3 = Gauss(mu=mu3, sigma=sigma3, obs=obs1, name="gauss3asum")
    gauss_dists = [gauss1, gauss2, gauss3]
    return gauss_dists


def sum_gaussians():
    gauss_dists = create_gaussians()
    sum_gauss = SumPDF(pdfs=gauss_dists, fracs=fracs, obs=obs1)
    return sum_gauss


def product_gaussians():
    gauss_dists = create_gaussians()
    prod_gauss = ProductPDF(pdfs=gauss_dists, obs=obs1)
    return prod_gauss


def product_gauss_3d(name=""):
    # Gauss for product, independent
    mu1, mu2, mu3, sigma1, sigma2, sigma3 = create_params('3d' + name)

    gauss13 = Gauss(mu=mu1, sigma=sigma1, obs=obs1, name="gauss1a")
    gauss23 = Gauss(mu=mu2, sigma=sigma2, obs=obs2, name="gauss2a")
    gauss33 = Gauss(mu=mu3, sigma=sigma3, obs=obs3, name="gauss3a")
    gauss_dists3 = [gauss13, gauss23, gauss33]
    prod_gauss_3d = ProductPDF(pdfs=gauss_dists3)
    return prod_gauss_3d

    # prod_gauss_3d.update_integration_options(draws_per_dim=33)


def product_gauss_4d():
    mu1, mu2, mu3, sigma1, sigma2, sigma3 = create_params("4d")

    obs4 = zfit.Space('obs4', (-4.5, 4.7))
    gauss12 = Gauss(mu=mu1, sigma=sigma1, obs=obs4, name="gauss12a")
    gauss22 = Gauss(mu=mu2, sigma=sigma2, obs=obs1, name="gauss22a")
    gauss32 = Gauss(mu=mu3, sigma=sigma3, obs=obs3, name="gauss32a")
    gauss_dists2 = [gauss12, gauss22, gauss32]
    obs = zfit.Space(['obs1', 'obs2', 'obs3', 'obs4'])

    prod_3d = product_gauss_3d("4d")
    prod_gauss_4d = ProductPDF(pdfs=gauss_dists2 + [prod_3d], obs=obs)
    return prod_gauss_4d


def test_prod_gauss_nd():
    test_values = np.random.random(size=(10, 3))
    test_values_data = Data.from_tensor(obs=obs1 * obs2 * obs3, tensor=test_values)
    product_pdf = product_gauss_3d()
    assert product_pdf.n_obs == 3
    probs = product_pdf.pdf(x=test_values_data)
    gaussians = create_gaussians()
    for gauss, space in zip(gaussians, [obs1, obs2, obs3]):
        gauss.set_norm_range(space.rect_limits)
    true_probs = np.prod(
        [gauss.pdf(test_values[:, i]) for i, gauss in enumerate(gaussians)], axis=0)
    probs_np = probs.numpy()
    np.testing.assert_allclose(true_probs, probs_np, rtol=1e-2)


@pytest.mark.flaky(reruns=3)
def test_prod_gauss_nd_mixed():
    norm_range = (-5, 4)
    low, high = norm_range
    test_values = np.random.uniform(low=low, high=high, size=(1000, 4))

    obs4d = ['obs1', 'obs2', 'obs3', 'obs4']
    test_values_data = Data.from_tensor(obs=obs4d, tensor=test_values)
    # prod_gauss_4d.update_integration_options(draws_per_dim=10)
    limits_4d = Space(limits=(((-5,) * 4,), ((4,) * 4,)), obs=obs4d)
    prod_gauss_4d = product_gauss_4d()
    prod_gauss_4d.set_norm_range(limits_4d)
    probs = prod_gauss_4d.pdf(x=test_values_data,
                              norm_range=limits_4d)
    gausses = create_gaussians()

    for gauss in gausses:
        gauss.set_norm_range(norm_range)
    gauss1, gauss2, gauss3 = gausses
    prod_gauss_3d = product_gauss_3d()

    def probs_4d(values):
        true_prob = [gauss1.pdf(values[:, 3])]
        true_prob += [gauss2.pdf(values[:, 0])]
        true_prob += [gauss3.pdf(values[:, 2])]
        true_prob += [prod_gauss_3d.pdf(values[:, 0:3], norm_range=Space(limits=(((-5,) * 3,), ((4,) * 3,)),
                                                                         obs=['obs1', 'obs2', 'obs3']))]
        return tf.math.reduce_prod(true_prob, axis=0)

    true_unnormalized_probs = probs_4d(values=test_values)

    normalization_probs = limits_4d.area() * probs_4d(tf.random.uniform(minval=low, maxval=high, shape=(40 ** 4, 4)))
    true_probs = true_unnormalized_probs / tf.reduce_mean(input_tensor=normalization_probs)
    probs_np = probs.numpy()
    true_probs_np = true_probs.numpy()
    assert np.average(probs_np * limits_4d.area()) == pytest.approx(1., rel=0.33)  # low n mc
    np.testing.assert_allclose(true_probs_np, probs_np, rtol=2e-2)


def test_func_sum():
    sum_gauss = sum_gaussians()
    test_values = np.random.uniform(low=-3, high=4, size=10)
    sum_gauss_as_func = sum_gauss.as_func(norm_range=(-10, 10))
    vals = sum_gauss_as_func.func(x=test_values)
    vals = zfit.run(vals)
    # test_sum = sum([g.func(test_values) for g in gauss_dists])
    np.testing.assert_allclose(vals, true_gaussian_sum(test_values), rtol=3e-2)  # MC integral


def test_normalization_sum_gauss():
    normalization_testing(sum_gaussians())


def test_normalization_sum_gauss_extended():
    test_yield = 109.
    sum_gauss_extended = sum_gaussians().create_extended(test_yield)
    normalization_testing(sum_gauss_extended)


def test_normalization_prod_gauss():
    normalization_testing(product_gaussians())


def test_exp():
    lambda_true = 0.031
    lambda_ = zfit.Parameter('lambda1', lambda_true)
    exp1 = zfit.pdf.Exponential(lam=lambda_, obs=zfit.Space('obs1', (-11, 11)))
    sample = exp1.sample(n=1000, limits=(-10, 10))
    sample_np = sample.numpy()
    assert not any(np.isnan(sample_np))

    exp2 = zfit.pdf.Exponential(lam=lambda_, obs=zfit.Space('obs1', (5250, 5750)))
    probs2 = exp2.pdf(x=np.linspace(5300, 5700, num=1100))
    probs2_np = probs2.numpy()
    assert not any(np.isnan(probs2_np))
    normalization_testing(exp2, limits=(5400, 5800))

    intlim = [5400, 5500]
    integral2 = zfit.run(exp2.integrate(intlim))
    numintegral2 = zfit.run(exp2.numeric_integrate(intlim))
    assert integral2 == pytest.approx(numintegral2, rel=0.03)


def normalization_testing(pdf, limits=None):
    limits = (low, high) if limits is None else limits
    space = Space(obs=obs1, limits=limits)
    with pdf.set_norm_range(space):
        samples = tf.cast(np.random.uniform(low=space.lower, high=space.upper, size=(40000, pdf.n_obs)),
                          dtype=tf.float64)
        samples = zfit.Data.from_tensor(obs=space, tensor=samples)
        probs = pdf.pdf(samples)
        result = probs.numpy()
        result = zfit.run(np.average(result) * space.rect_area())
        assert pytest.approx(result, rel=0.03) == 1


def test_extended_gauss():
    mu1 = Parameter("mu11", 1.)
    mu2 = Parameter("mu21", 12.)
    mu3 = Parameter("mu31", 3.)
    sigma1 = Parameter("sigma11", 1.)
    sigma2 = Parameter("sigma21", 12.)
    sigma3 = Parameter("sigma31", 33.)
    yield1 = Parameter("yield11", 150.)
    yield2 = Parameter("yield21", 550.)
    yield3 = Parameter("yield31", 2500.)

    gauss1 = Gauss(mu=mu1, sigma=sigma1, obs=obs1, name="gauss11")
    gauss2 = Gauss(mu=mu2, sigma=sigma2, obs=obs1, name="gauss21")
    gauss3 = Gauss(mu=mu3, sigma=sigma3, obs=obs1, name="gauss31")
    gauss1 = gauss1.create_extended(yield1)
    gauss2 = gauss2.create_extended(yield2)
    gauss3 = gauss3.create_extended(yield3)

    gauss_dists = [gauss1, gauss2, gauss3]

    sum_gauss = SumPDF(pdfs=gauss_dists)
    integral_true = sum_gauss.integrate((-1, 5)) * sum_gauss.get_yield()

    assert zfit.run(integral_true) == pytest.approx(zfit.run(sum_gauss.ext_integrate((-1, 5))))
    normalization_testing(pdf=sum_gauss, limits=obs1)


if __name__ == '__main__':
    test_prod_gauss_nd()



#!/usr/bin/env python
#  Copyright (c) 2021 zfit

import argparse
import os
import re

import yaml

here = os.path.dirname(os.path.realpath(__file__))

parser = argparse.ArgumentParser(
    description='Apply cuts to ntuple',
    formatter_class=argparse.ArgumentDefaultsHelpFormatter
)

parser.add_argument("files", nargs='*',
                    help='Files to be processed.')

parser.add_argument("--dry", action="store_true",
                    help='Dry run WITHOUT replacing.')

cfg = parser.parse_args()

with open(here + '/argdocs.yaml') as replfile:
    replacements = yaml.load(replfile, Loader=yaml.Loader)

# Replace the target string
# auto_end_old = r'|@docend|'
for filepath in cfg.files:
    if not filepath.endswith('.py'):
        continue
    with open(filepath) as file:
        filedata = file.read()

    infile = False
    needs_replacement = False
    for param, replacement in replacements.items():
        replacement = replacement.rstrip('\n')
        while replacement[:1] == ' ':  # we want to remove the whitespace
            replacement = replacement[1:]
        auto_start = fr'|@doc:{param}|'
        auto_end = fr'|@docend:{param}|'
        matches = re.findall(auto_start.replace('|', r'\|')
                             + r".*?"
                             + auto_end.replace('|', r'\|'), filedata, re.DOTALL)

        if not matches:
            continue
        infile = True

        replacement_mod = f'{auto_start} {replacement} {auto_end}'

        for match in matches:
            if auto_start in match[len(auto_start):]:  # sanity check
                raise ValueError(f"Docstring formatting error,"
                                 f" has more than one start until an end command: {match}")
            if match != replacement_mod:
                needs_replacement = True
                filedata = filedata.replace(match, replacement_mod)

    # Write the file out again
    replace_msg = 'replaced docs' if needs_replacement else 'docs already there'
    filename = filepath.split("/")[-1]
    if infile:
        if cfg.dry:
            print(f'Match in {filename}, {replace_msg}, not writing to {filepath}, dry run.')
        else:
            if needs_replacement:
                with open(filepath, 'w') as file:
                    file.write(filedata)
                print(f'Written to {filename}.')

            else:
                print(f'Match in {filename}, {replace_msg}, not writing as not needed')



#  Copyright (c) 2021 zfit
import pytest
import tensorflow as tf

import zfit
from zfit import z
from zfit.core.space import ANY, Limit

rect_limits = (1., 3)
rect_limits_tf = (tf.constant(1.), tf.constant(3))
rect_limit_enlarged = rect_limits[0] - 1, rect_limits[1] + 1.5
rect_limits_any = (ANY, ANY)


def inside_rect_limits(x):
    return tf.logical_and(rect_limits[0] < x, x < rect_limits[1])[:, 0]


@pytest.mark.parametrize('graph', [True, False])
@pytest.mark.parametrize('rect_limits', [rect_limits, rect_limits_tf])
@pytest.mark.parametrize('testclass',
                         [Limit, lambda limit_fn=None, rect_limits=None: zfit.Space('obs1', limits=limit_fn,
                                                                                    rect_limits=rect_limits)])
def test_rect_limits_1d(graph, rect_limits, testclass):
    def test(allow_graph=True):
        limit = testclass(rect_limits)
        limit2 = testclass(rect_limits=rect_limits)

        lower, upper = limit.rect_limits
        lower2, upper2 = limit2.rect_limits
        assert limit.has_rect_limits
        inside = limit.inside(2)
        inside2 = limit.inside(2, guarantee_limits=True)
        outside = limit.inside(4)
        if graph:
            equal = limit.equal(limit2, allow_graph=allow_graph)
            less_equal = limit.less_equal(limit2, allow_graph=allow_graph)
        else:
            equal = limit == limit2
            less_equal = limit <= limit2

        return lower, upper, lower2, upper2, inside, inside2, outside, equal, less_equal

    if graph:
        test = z.function(test)
    lower, upper, lower2, upper2, inside, inside2, outside, equal, less_equal = test()
    assert equal
    assert less_equal
    assert lower, upper == rect_limits
    assert lower2, upper2 == rect_limits
    assert inside
    assert inside2
    assert not outside


@pytest.mark.parametrize('graph', [True, False])
@pytest.mark.parametrize('limit_fn', [None, inside_rect_limits])
@pytest.mark.parametrize('testclass', [Limit, lambda limit_fn, rect_limits: zfit.Space('obs1', limits=limit_fn,
                                                                                       rect_limits=rect_limits)])
def test_less_equal(graph, limit_fn, testclass):
    def test(limit_fn=None, allow_graph=True):
        limit = testclass(limit_fn=limit_fn, rect_limits=rect_limits)
        limit2 = testclass(limit_fn=limit_fn, rect_limits=rect_limit_enlarged)

        assert limit.has_rect_limits ^ bool(limit_fn)
        assert limit.has_limits
        assert limit.limits_are_set
        assert not limit.limits_are_false
        inside = limit.inside(2)
        inside2 = limit.inside(2, guarantee_limits=True)
        outside = limit.inside(4)

        assert limit2.has_rect_limits ^ bool(limit_fn)
        assert limit2.has_limits
        assert limit2.limits_are_set
        assert not limit2.limits_are_false
        inside21 = limit2.inside(2)
        inside22 = limit2.inside(2, guarantee_limits=True)
        outside2 = limit2.inside(4)
        if graph:
            equal = limit.equal(limit2, allow_graph=allow_graph)
        else:
            equal = limit == limit2

        return inside, inside2, outside, equal, inside21, inside22, outside2

    if graph:
        test = z.function(test)
    inside, inside2, outside, equal, inside21, inside22, outside2 = test(limit_fn=limit_fn)
    assert not (equal ^ bool(limit_fn))  # if a limit_fn is specified, this has precedency over the rect
    assert inside
    assert inside2
    assert not outside
    assert inside21
    assert inside22
    assert bool(outside2) ^ bool(limit_fn)  # if limit_fn, this is outside


@pytest.mark.parametrize('graph', [True, False])
@pytest.mark.parametrize('testclass', [Limit, lambda limit_fn, rect_limits: zfit.Space('obs1', limits=limit_fn,
                                                                                       rect_limits=rect_limits)])
def test_limits_1d(graph, testclass):
    def test(allow_graph=True):
        limit = testclass(limit_fn=inside_rect_limits, rect_limits=rect_limits)
        limit2 = testclass(limit_fn=inside_rect_limits, rect_limits=rect_limit_enlarged)

        assert not limit.has_rect_limits
        assert limit.has_limits
        assert limit.limits_are_set
        assert not limit.limits_are_false
        inside = limit.inside(2)
        inside2 = limit.inside(2, guarantee_limits=True)
        outside = limit.inside(4)

        assert not limit2.has_rect_limits
        assert limit2.has_limits
        assert limit2.limits_are_set
        assert not limit2.limits_are_false
        inside21 = limit2.inside(2)
        inside22 = limit2.inside(2, guarantee_limits=True)
        outside2 = limit2.inside(4)
        if graph:
            equal = limit.equal(limit2, allow_graph=allow_graph)
        else:
            equal = limit == limit2

        return inside, inside2, outside, equal, inside21, inside22, outside2

    if graph:
        test = z.function(test)
    inside, inside2, outside, equal, inside21, inside22, outside2 = test()
    assert equal
    assert inside
    assert inside2
    assert not outside
    assert inside21
    assert inside22
    assert not outside2



#  Copyright (c) 2021 zfit


from .minimizers.baseminimizer import (BaseMinimizer, BaseMinimizerV1,
                                       DefaultStrategy, minimize_supports)
from .minimizers.ipopt import IpyoptV1
from .minimizers.minimizer_minuit import Minuit
from .minimizers.minimizer_nlopt import (NLoptBaseMinimizerV1, NLoptBOBYQAV1,
                                         NLoptCCSAQV1, NLoptCOBYLAV1,
                                         NLoptESCHV1, NLoptISRESV1,
                                         NLoptLBFGSV1, NLoptMLSLV1, NLoptMMAV1,
                                         NLoptShiftVarV1, NLoptSLSQPV1,
                                         NLoptStoGOV1, NLoptSubplexV1,
                                         NLoptTruncNewtonV1)
from .minimizers.minimizers_scipy import (ScipyBaseMinimizerV1, ScipyLBFGSBV1,
                                          ScipyNelderMeadV1, ScipyNewtonCGV1,
                                          ScipyPowellV1, ScipySLSQPV1,
                                          ScipyTruncNCV1, ScipyTrustConstrV1)
from .minimizers.optimizers_tf import Adam, WrapOptimizer
from .minimizers.strategy import (DefaultToyStrategy, PushbackStrategy,
                                  ZfitStrategy)
from .minimizers.termination import EDM
from .util.deprecation import deprecated


class AdamMinimizer(Adam):

    @deprecated(None, "Use zfit.minimize.Adam instead.")
    def __init__(self, tol=None, learning_rate=0.2, beta1=0.9, beta2=0.999, epsilon=1e-08, name='Adam', **kwargs):
        super().__init__(tol, learning_rate, beta1, beta2, epsilon, name, **kwargs)


class MinuitMinimizer(Minuit):

    @deprecated(None, "Use zfit.minimize.Minuit instead.")
    def __init__(self, strategy: ZfitStrategy = None, minimize_strategy: int = 1, tol: float = None,
                 verbosity: int = 5, name: str = None, ncall: int = 10000, use_minuit_grad: bool = None,
                 **options):
        super().__init__(strategy, minimize_strategy, tol, verbosity, name, ncall, use_minuit_grad,
                         **options)


class Scipy:
    def __init__(self, *_, **__):
        raise RuntimeError("This has been removed. Use the new Scipy* minimizer instead. In case anyone is missing,"
                           " please feel free to open a request:"
                           " https://github.com/zfit/zfit/issues/new?assignees="
                           "&labels=discussion&template=feature-request.md&title="
                           " or directly make a PR.")


class ScipyMinimizer(Scipy):
    pass


class BFGS:

    def __init__(self) -> None:
        raise RuntimeError("BFGS (from TensorFlow Probability) has been removed as it is currently"
                           " not working well. Use other BFGS-like implementations such as ScipyLBFGSBV1"
                           " or NLoptLBFGSV1.")


__all__ = ["WrapOptimizer",
           "Adam", "Minuit",
           "ScipyBaseMinimizerV1",
           "ScipyLBFGSBV1", 'ScipyTrustConstrV1',
           "ScipyPowellV1", "ScipySLSQPV1", "ScipyNewtonCGV1", "ScipyTruncNCV1", "ScipyNelderMeadV1",
           "NLoptBaseMinimizerV1",
           "NLoptLBFGSV1", "NLoptTruncNewtonV1", "NLoptSLSQPV1", "NLoptMMAV1", "NLoptCCSAQV1", 'NLoptShiftVarV1',
           "NLoptMLSLV1", 'NLoptStoGOV1', 'NLoptESCHV1', 'NLoptISRESV1', "NLoptSubplexV1", "NLoptBOBYQAV1",
           'NLoptCOBYLAV1',
           "IpyoptV1",
           "BaseMinimizer", "BaseMinimizerV1", "minimize_supports",
           "DefaultStrategy", "DefaultToyStrategy", "PushbackStrategy",
           "EDM"]



#  Copyright (c) 2021 zfit
import numpy as np
import pytest
import tensorflow as tf

import zfit
from zfit import Space, z
from zfit.core.space import Limit
from zfit.util.exception import AnalyticSamplingNotImplemented


@pytest.fixture(autouse=True, scope="module")
def setup_teardown_vectors():
    Limit._experimental_allow_vectors = True
    yield
    Limit._experimental_allow_vectors = False


mu_true = 1.5
sigma_true = 1.2
low, high = -3.8, 2.9

obs1 = zfit.Space('obs1', (low, high))


class GaussNoAnalyticSampling(zfit.pdf.Gauss):

    def _analytic_sample(self, n, limits: Space):
        raise AnalyticSamplingNotImplemented  # HACK do make importance sampling work


class UniformNoAnalyticSampling(zfit.pdf.Uniform):

    def _analytic_sample(self, n, limits: Space):
        raise AnalyticSamplingNotImplemented  # HACK do make importance sampling work


def create_gauss1():
    mu = zfit.Parameter("mu_sampling1", mu_true, mu_true - 2., mu_true + 7.)
    sigma = zfit.Parameter("sigma_sampling1", sigma_true, sigma_true - 10., sigma_true + 5.)

    gauss_params1 = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs1, name="gauss_params1_sampling1")
    return gauss_params1, mu, sigma


class TmpGaussian(zfit.pdf.BasePDF):

    def __init__(self, obs, mu, sigma, params=None,
                 name: str = "BasePDF", **kwargs):
        params = {'mu': mu, 'sigma': sigma}
        super().__init__(obs, params, name=name, **kwargs)

    def _unnormalized_pdf(self, x):
        x = x.unstack_x()
        mu = self.params['mu']
        sigma = self.params['sigma']

        return z.exp((-(x - mu) ** 2) / (
            2 * sigma ** 2))  # non-normalized gaussian


class TmpGaussianPDFNonNormed(zfit.pdf.BasePDF):

    def __init__(self, obs, mu, sigma, params=None,
                 name: str = "BasePDF", **kwargs):
        params = {'mu': mu, 'sigma': sigma}
        super().__init__(obs, params, name=name, **kwargs)

    def _pdf(self, x, norm_range=False):
        x = x.unstack_x()
        mu = self.params['mu']
        sigma = self.params['sigma']

        return z.exp((-(x - mu) ** 2) / (
            2 * sigma ** 2))  # non-normalized gaussian


def create_test_gauss1():
    mu2 = zfit.Parameter("mu2_sampling1", mu_true, mu_true - 2., mu_true + 7.)
    sigma2 = zfit.Parameter("sigma2_sampling1", sigma_true, sigma_true - 10., sigma_true + 5.)

    test_gauss1 = TmpGaussian(name="test_gauss1", mu=mu2, sigma=sigma2, obs=obs1)
    return test_gauss1, mu2, sigma2


def create_test_pdf_overriden_gauss1():
    mu2 = zfit.Parameter("mu2_sampling1", mu_true, mu_true - 2., mu_true + 7.)
    sigma2 = zfit.Parameter("sigma2_sampling1", sigma_true, sigma_true - 10., sigma_true + 5.)

    test_gauss1 = TmpGaussianPDFNonNormed(name="test_pdf_nonnormed_gauss1", mu=mu2, sigma=sigma2, obs=obs1)
    return test_gauss1, mu2, sigma2


gaussian_dists = [create_gauss1, create_test_gauss1]


def test_mutlidim_sampling():
    spaces = [zfit.Space(f'obs{i}', (i * 10, i * 10 + 6)) for i in range(4)]
    pdfs = [GaussNoAnalyticSampling(obs=spaces[0], mu=3, sigma=1),
            GaussNoAnalyticSampling(obs=spaces[2], mu=23, sigma=1),
            UniformNoAnalyticSampling(obs=spaces[1], low=12, high=14),
            UniformNoAnalyticSampling(obs=spaces[3], low=32, high=34),
            ]
    prod = zfit.pdf.ProductPDF(pdfs)
    sample = prod.sample(n=20000)
    for i, space in enumerate([p.space for p in pdfs]):
        assert all(space.inside(sample.value()[:, i]))


@pytest.mark.flaky(2)  # sampling
@pytest.mark.parametrize('gauss_factory', gaussian_dists)
def test_multiple_limits_sampling(gauss_factory):
    gauss, mu, sigma = gauss_factory()

    low = mu_true - 6 * sigma_true
    high = mu_true + 6 * sigma_true
    low1 = low
    low2 = up1 = (high - low) / 5
    up2 = high
    obs11 = zfit.Space(obs1.obs, (low1, up1))
    obs12 = zfit.Space(obs1.obs, (low2, up2))
    obs_split = obs11 + obs12
    obs = zfit.Space(obs1.obs, (low, high))

    n = 50000
    sample1 = gauss.sample(n=n, limits=obs)
    sample2 = gauss.sample(n=n, limits=obs_split)

    rel_tol = 1e-2
    assert np.mean(sample1.value()) == pytest.approx(mu_true, rel_tol)
    assert np.std(sample1.value()) == pytest.approx(sigma_true, rel_tol)
    assert np.mean(sample2.value()) == pytest.approx(mu_true, rel_tol)
    assert np.std(sample2.value()) == pytest.approx(sigma_true, rel_tol)


@pytest.mark.parametrize('gauss_factory', gaussian_dists + [create_test_pdf_overriden_gauss1])
def test_sampling_fixed(gauss_factory):
    gauss, mu, sigma = gauss_factory()

    n_draws = 1000
    n_draws_param = tf.Variable(initial_value=n_draws, trainable=False, dtype=tf.int64,
                                name='n_draws')  # variable to have something changeable, predictable
    sample_tensor = gauss.create_sampler(n=n_draws_param, limits=(low, high))
    sample_tensor.resample()
    sampled_from_gauss1 = sample_tensor.numpy()
    assert max(sampled_from_gauss1[:, 0]) <= high
    assert min(sampled_from_gauss1[:, 0]) >= low
    assert n_draws == len(sampled_from_gauss1[:, 0])

    new_n_draws = 867
    n_draws_param.assign(new_n_draws)
    sample_tensor.resample()
    sampled_from_gauss1_small = sample_tensor.numpy()
    assert new_n_draws == len(sampled_from_gauss1_small[:, 0])
    assert not np.allclose(sampled_from_gauss1[:new_n_draws], sampled_from_gauss1_small)
    n_draws_param.assign(n_draws)

    gauss_full_sample = gauss.create_sampler(n=10000,
                                             limits=(mu_true - abs(sigma_true) * 3, mu_true + abs(sigma_true) * 3))
    gauss_full_sample.resample()
    sampled_gauss1_full = gauss_full_sample.numpy()
    mu_sampled = np.mean(sampled_gauss1_full)
    sigma_sampled = np.std(sampled_gauss1_full)
    assert mu_sampled == pytest.approx(mu_true, rel=0.07)
    assert sigma_sampled == pytest.approx(sigma_true, rel=0.07)

    with mu.set_value(mu_true - 1):
        sample_tensor.resample()
        sampled_from_gauss1 = sample_tensor.numpy()
        assert max(sampled_from_gauss1[:, 0]) <= high
        assert min(sampled_from_gauss1[:, 0]) >= low
        assert n_draws == len(sampled_from_gauss1[:, 0])

        sampled_gauss1_full = gauss_full_sample.numpy()
        mu_sampled = np.mean(sampled_gauss1_full)
        sigma_sampled = np.std(sampled_gauss1_full)
        assert mu_sampled == pytest.approx(mu_true, rel=0.07)
        assert sigma_sampled == pytest.approx(sigma_true, rel=0.07)

    gauss_full_sample2 = gauss.create_sampler(n=10000, limits=(-10, 10))

    gauss_full_sample2.resample(param_values={mu: mu_true - 1.0})
    sampled_gauss2_full = gauss_full_sample2.numpy()
    mu_sampled = np.mean(sampled_gauss2_full)
    sigma_sampled = np.std(sampled_gauss2_full)
    assert mu_sampled == pytest.approx(mu_true - 1.0, rel=0.08)
    assert sigma_sampled == pytest.approx(sigma_true, rel=0.08)

    gauss_full_sample2.resample(param_values={sigma: sigma_true + 1.0})
    sampled_gauss2_full = gauss_full_sample2.numpy()
    mu_sampled = np.mean(sampled_gauss2_full)
    sigma_sampled = np.std(sampled_gauss2_full)
    assert mu_sampled == pytest.approx(mu_true, rel=0.07)
    assert sigma_sampled == pytest.approx(sigma_true + 1.0, rel=0.07)


@pytest.mark.parametrize('gauss_factory', gaussian_dists)
def test_sampling_floating(gauss_factory):
    gauss, mu, sigma = gauss_factory()

    n_draws = 1000
    sampler = gauss.create_sampler(n=n_draws, limits=(low, high), fixed_params=False)
    sampler.resample()
    sampled_from_gauss1 = sampler
    assert max(sampled_from_gauss1[:, 0]) <= high
    assert min(sampled_from_gauss1[:, 0]) >= low
    assert n_draws == len(sampled_from_gauss1[:, 0])

    gauss_full_sample = gauss.create_sampler(n=10000,
                                             limits=(mu_true - abs(sigma_true) * 3, mu_true + abs(sigma_true) * 3),
                                             fixed_params=False)

    gauss_full_sample_fixed = gauss.create_sampler(n=10000,
                                                   limits=(
                                                       mu_true - abs(sigma_true) * 3, mu_true + abs(sigma_true) * 3),
                                                   fixed_params=True)
    gauss_full_sample.resample()
    gauss_full_sample_fixed.resample()
    assert set(gauss_full_sample_fixed.fixed_params) == {mu, sigma}
    sampled_gauss1_full = gauss_full_sample.numpy()
    mu_sampled = np.mean(sampled_gauss1_full)
    sigma_sampled = np.std(sampled_gauss1_full)
    assert mu_sampled == pytest.approx(mu_true, rel=0.07)
    assert sigma_sampled == pytest.approx(sigma_true, rel=0.07)

    sampled_gauss1_full_fixed = gauss_full_sample_fixed.numpy()
    mu_sampled_fixed = np.mean(sampled_gauss1_full_fixed)
    sigma_sampled_fixed = np.std(sampled_gauss1_full_fixed)
    assert mu_sampled_fixed == pytest.approx(mu_true, rel=0.07)
    assert sigma_sampled_fixed == pytest.approx(sigma_true, rel=0.07)

    mu_diff = 0.7
    with mu.set_value(mu_true - mu_diff):
        assert mu.numpy() == mu_true - mu_diff
        sampler.resample()
        sampled_from_gauss1 = sampler.numpy()
        assert max(sampled_from_gauss1[:, 0]) <= high
        assert min(sampled_from_gauss1[:, 0]) >= low
        assert n_draws == len(sampled_from_gauss1[:, 0])

        sampled_gauss1_full_fixed = gauss_full_sample_fixed.numpy()
        mu_sampled_fixed = np.mean(sampled_gauss1_full_fixed)
        sigma_sampled_fixed = np.std(sampled_gauss1_full_fixed)
        assert mu_sampled_fixed == pytest.approx(mu_true, rel=0.07)
        assert sigma_sampled_fixed == pytest.approx(sigma_true, rel=0.07)

        gauss_full_sample.resample()
        sampled_gauss1_full = gauss_full_sample.numpy()
        mu_sampled = np.mean(sampled_gauss1_full)
        sigma_sampled = np.std(sampled_gauss1_full)
        assert mu_sampled == pytest.approx(mu_true - mu_diff, rel=0.07)
        assert sigma_sampled == pytest.approx(sigma_true, rel=0.07)


# @pytest.mark.skipif(not zfit.EXPERIMENTAL_FUNCTIONS_RUN_EAGERLY, reason="deadlock in tf.function, issue #35540")  # currently, importance sampling is not working, odd deadlock in TF
@pytest.mark.flaky(3)  # statistical
def test_importance_sampling():
    from zfit.core.sample import accept_reject_sample

    mu_sampler = 5.
    sigma_sampler = 4.
    mu_pdf = 4.
    sigma_pdf = 1.

    obs_sampler = zfit.Space(obs='obs1', limits=(4.5, 5.5))  # smaller, so pdf is bigger
    obs_pdf = zfit.Space(obs='obs1', limits=(1, 7))

    gauss_sampler = GaussNoAnalyticSampling(mu=mu_sampler, sigma=sigma_sampler, obs=obs_sampler)
    gauss_pdf = GaussNoAnalyticSampling(mu=mu_pdf, sigma=sigma_pdf, obs=obs_pdf)

    importance_sampling_called = [False]

    class GaussianSampleAndWeights:

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        @z.function
        def __call__(self, n_to_produce, limits, dtype):
            importance_sampling_called[0] = True
            n_to_produce = tf.cast(n_to_produce, dtype=tf.int32)
            gaussian_sample = gauss_sampler.sample(n=n_to_produce, limits=limits)
            weights = gauss_sampler.pdf(gaussian_sample)
            weights_max = None
            thresholds = tf.random.uniform(shape=(n_to_produce,), dtype=dtype)
            return gaussian_sample, thresholds, weights, weights_max, n_to_produce

    sample = accept_reject_sample(prob=lambda x: gauss_pdf.pdf(x, norm_range=False), n=30000, limits=obs_pdf)
    gauss_pdf._sample_and_weights = GaussianSampleAndWeights
    sample2 = gauss_pdf.sample(n=30000, limits=obs_pdf)
    assert importance_sampling_called[0]
    sample_np, sample_np2 = [sample.numpy(), sample2.numpy()]

    mean = np.mean(sample_np)
    mean2 = np.mean(sample_np2)
    std = np.std(sample_np)
    std2 = np.std(sample_np2)
    assert mean == pytest.approx(mu_pdf, rel=0.02)
    assert mean2 == pytest.approx(mu_pdf, rel=0.02)
    assert std == pytest.approx(sigma_pdf, rel=0.02)
    assert std2 == pytest.approx(sigma_pdf, rel=0.02)


@pytest.mark.flaky(3)  # statistical
def test_importance_sampling_uniform():
    low = -3.
    high = 7.
    obs = zfit.Space("obs1", (low, high))
    uniform = UniformNoAnalyticSampling(obs=obs, low=low, high=high)
    importance_sampling_called = [False]

    class GaussianSampleAndWeights:

        def __call__(self, n_to_produce, limits, dtype):
            importance_sampling_called[0] = True

            import tensorflow_probability.python.distributions as tfd
            n_to_produce = tf.cast(n_to_produce, dtype=tf.int32)
            gaussian = tfd.TruncatedNormal(loc=z.constant(-1.), scale=z.constant(2.),
                                           low=low, high=high)
            sample = gaussian.sample(sample_shape=(n_to_produce, 1))
            weights = gaussian.prob(sample)[:, 0]
            thresholds = tf.random.uniform(shape=(n_to_produce,), dtype=dtype)
            return sample, thresholds, weights, None, n_to_produce

    uniform._sample_and_weights = GaussianSampleAndWeights
    n_sample = 10000
    sample = uniform.sample(n=n_sample)
    assert importance_sampling_called[0]
    sample_np = sample.numpy()
    n_bins = 20
    bin_counts, _ = np.histogram(sample_np, bins=n_bins)
    expected_per_bin = n_sample / n_bins

    assert np.std(bin_counts) < np.sqrt(expected_per_bin) * 2
    assert all(abs(bin_counts - expected_per_bin) < np.sqrt(expected_per_bin) * 5)
    # import matplotlib.pyplot as plt
    # plt.hist(sample_np, bins=40)
    # plt.show()


# @pytest.mark.xfail  # TODO(mayou36): hacked, EventSpace, needed for phasespace sampling
def test_sampling_fixed_eventlimits():
    n_samples1 = 500
    n_samples2 = 400  # just to make sure
    n_samples3 = 356  # just to make sure
    n_samples_tot = n_samples1 + n_samples2 + n_samples3

    obs1 = "obs1"
    lower1, upper1 = -10, -9
    lower2, upper2 = 0, 1
    lower3, upper3 = 10, 11
    lower = tf.convert_to_tensor(value=tuple([lower1] * n_samples1
                                             + [lower2] * n_samples2
                                             + [lower3] * n_samples3))[:, None]
    upper = tf.convert_to_tensor(value=tuple([upper1] * n_samples1
                                             + [upper2] * n_samples2
                                             + [upper3] * n_samples3))[:, None]

    # limits = zfit.core.sample.EventSpace(obs=obs1, limits=(lower, upper))
    limits = zfit.Space(obs=obs1, limits=(lower, upper))
    gauss1 = GaussNoAnalyticSampling(mu=0.3, sigma=4, obs=zfit.Space(obs=obs1, limits=(-12, 12)))

    sample = gauss1.sample(n=n_samples_tot, limits=limits)
    sample_np = sample.numpy()
    assert sample_np.shape[0] == n_samples_tot
    assert all(lower1 <= sample_np[:n_samples1])
    assert all(sample_np[:n_samples1] <= upper1)
    assert all(lower2 <= sample_np[n_samples1:n_samples2])
    assert all(sample_np[n_samples1:n_samples2] <= upper2)
    assert all(lower3 <= sample_np[n_samples2:n_samples3])
    assert all(sample_np[n_samples2:n_samples3] <= upper3)
    # with pytest.raises(InvalidArgumentError):  # cannot use the exact message, () are regex syntax... bug in pytest
    #     _ = gauss1.sample(n=n_samples_tot + 1, limits=limits)  # TODO(Mayou36): catch analytic integral



#  Copyright (c) 2021 zfit
from typing import Union

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability.python import distributions as tfd

import zfit.z.numpy as znp

from .. import z
from ..core.interfaces import ZfitData, ZfitSpace
from ..settings import ztypes
from ..util import ztyping
from ..util.exception import OverdefinedError, ShapeIncompatibleError
from .dist_tfp import WrapDistribution


def bandwidth_rule_of_thumb(data, factor=0.9):
    return min_std_or_iqr(data) * tf.cast(tf.shape(data)[0], ztypes.float) ** (-1 / 5.) * factor


def bandwidth_silverman(data):
    return bandwidth_rule_of_thumb(data=data, factor=0.9)


def bandwidth_scott(data):
    return bandwidth_rule_of_thumb(data=data, factor=1.059)


def bandwidth_adaptiveV1(data, func):
    from .. import run
    run.assert_executing_eagerly()
    data = z.convert_to_tensor(data)
    bandwidth = z.sqrt(min_std_or_iqr(data) / func(data))
    bandwidth *= tf.cast(tf.shape(data)[0], ztypes.float) ** (-1 / 5.) * 1.059
    return bandwidth


def _adaptive_bandwidth_KDEV1(constructor, obs, data, weights, name, truncate):
    kde_silverman = constructor(obs=obs, data=data, bandwidth='silverman', weights=weights,
                                name=f"INTERNAL_{name}", truncate=truncate)
    return bandwidth_adaptiveV1(data=data, func=kde_silverman.pdf)


def _bandwidth_scott_KDEV1(data, *_, **__):
    return bandwidth_scott(data)


def _bandwidth_silverman_KDEV1(data, *_, **__):
    return bandwidth_silverman(data)


@z.function(wraps='tensor')
def min_std_or_iqr(x):
    return znp.minimum(znp.std(x), (tfp.stats.percentile(x, 75) - tfp.stats.percentile(x, 25)))


class GaussianKDE1DimV1(WrapDistribution):
    _N_OBS = 1

    _bandwidth_methods = {
        'scott': _bandwidth_scott_KDEV1,
        'silverman': _bandwidth_silverman_KDEV1,
        'adaptiveV1': _adaptive_bandwidth_KDEV1,
        'adaptive': _adaptive_bandwidth_KDEV1,
    }

    def __init__(self, obs: ztyping.ObsTypeInput, data: ztyping.ParamTypeInput,
                 bandwidth: Union[ztyping.ParamTypeInput, str] = None,
                 weights: Union[None, np.ndarray, tf.Tensor] = None, truncate: bool = False,
                 name: str = "GaussianKDE1DimV1"):
        r"""EXPERIMENTAL, `FEEDBACK WELCOME <https://github.com/zfit/zfit/issues/new?assignees=&labels=&template=other.md&title=>`_
        One dimensional, (truncated) Kernel Density Estimation with a Gaussian Kernel.

        Kernel Density Estimation is a non-parametric method to approximate the density of given points.

        .. math::

            f_h(x) =  \frac{1}{nh} \sum_{i=1}^n K\Big(\frac{x-x_i}{h}\Big)

        where the kernel in this case is a (truncated) Gaussian

        .. math::
            K = \exp \Big(\frac{(x - x_i)^2}{\sigma^2}\Big)


        The bandwidth of the kernel can be estimated in different ways. It can either be a global bandwidth,
        corresponding to a single value, or a local bandwidth, each corresponding to one data point.

        **Usage**

        The KDE can be instantiated by using a numpy-like data sample, preferably a `zfit.Data` object. This
        will be used as the mu of the kernels. The bandwidth can either be given as a parameter (with length
        1 for a global one or length equal to the data size) - a rather advanced concept for methods such as
        cross validation - or determined from the data automatically, either through a simple method like
        scott or silverman rule of thumbs or through an iterative, adaptive method.

        Examples
        --------

        .. code-block:: python

            # generate some example kernels
            size = 150
            data = np.random.normal(size=size, loc=2, scale=3)

            limits = (-15, 5)
            obs = zfit.Space("obs1", limits=limits)
            kde_silverman = zfit.pdf.GaussianKDE1DimV1(data=data, obs=obs)

            # for a better smoothing of the kernels, use an adaptive approach
            kde = zfit.pdf.GaussianKDE1DimV1(data=data, obs=obs, bandwidth='adaptive')


        Args:
            data: 1-D Tensor-like. The positions of the `kernel`, the :math:`x_i`. Determines how many kernels will be created.
            bandwidth: Bandwidth of the kernel. Valid options are {'silverman', 'scott', 'adaptive'} or a numerical.
                If a numerical is given, it as to be broadcastable to the batch and event shape of the distribution.
                A scalar or a `zfit.Parameter` will simply broadcast to `data` for a 1-D distribution.

            obs: Observables
            weights: Weights of each `data`, can be None or Tensor-like with shape compatible with `data`
            truncate: If a truncated Gaussian kernel should be used with the limits given by the `obs` lower and
                upper limits. This can cause NaNs in case datapoints are outside of the limits.
            name: Name of the PDF
        """
        if bandwidth is None:
            bandwidth = 'silverman'

        original_data = data

        if isinstance(data, ZfitData):
            if data.weights is not None:
                if weights is not None:
                    raise OverdefinedError("Cannot specify weights and use a `ZfitData` with weights.")
                else:
                    weights = data.weights

            if data.n_obs > 1:
                raise ShapeIncompatibleError(f"KDE is 1 dimensional, but data {data} has {data.n_obs} observables.")
            data = z.unstack_x(data)

        # create fraction for the sum
        shape_data = tf.shape(data)
        size = tf.cast(shape_data[0], dtype=ztypes.float)
        if weights is not None:
            probs = weights / znp.sum(weights)
        else:
            probs = tf.broadcast_to(1 / size, shape=(tf.cast(size, tf.int32),))
        categorical = tfd.Categorical(probs=probs)  # no grad -> no need to recreate

        # estimate bandwidth
        bandwidth_param = bandwidth
        if isinstance(bandwidth, str):
            bw_method = self._bandwidth_methods.get(bandwidth)
            if bw_method is None:
                raise ValueError(f"Cannot use {bandwidth} as a bandwidth method. Use a numerical value or one of"
                                 f" the defined methods: {list(self._bandwidth_methods.keys())}")
            bandwidth = bw_method(constructor=type(self), obs=obs, data=data, weights=weights,
                                  name=f"INTERNAL_{name}", truncate=truncate)

        bandwidth_param = -999 if bandwidth_param in (
        'adaptiveV1', 'adaptive') else bandwidth  # TODO: multiparam for bandwidth?

        params = {'bandwidth': bandwidth_param}

        # create distribution factory
        if truncate:
            if not isinstance(obs, ZfitSpace):
                raise ValueError(f"`obs` has to be a `ZfitSpace` if `truncated` is True.")
            inside = obs.inside(data)
            all_inside = znp.all(inside)
            tf.debugging.assert_equal(all_inside, True, message="Not all data points are inside the limits but"
                                                                " a truncate kernel was chosen.")

            def kernel_factory():
                return tfp.distributions.TruncatedNormal(loc=self._data, scale=self._bandwidth,
                                                         low=self.space.rect_lower,
                                                         high=self.space.rect_upper)
        else:
            def kernel_factory():
                return tfp.distributions.Normal(loc=self._data, scale=self._bandwidth)

        dist_kwargs = lambda: dict(mixture_distribution=categorical,
                                   components_distribution=kernel_factory())
        distribution = tfd.MixtureSameFamily

        super().__init__(obs=obs,
                         params=params,
                         dist_params={},
                         dist_kwargs=dist_kwargs,
                         distribution=distribution,
                         name=name)

        self._data_weights = weights
        self._bandwidth = bandwidth
        self._data = data
        self._original_data = original_data
        self._truncate = truncate



#  Copyright (c) 2021 zfit


# deactivating CUDA capable gpus
from zfit.z.tools import _auto_upcast

suppress_gpu = False
if suppress_gpu:
    import os

    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"  # see issue #152
    os.environ["CUDA_VISIBLE_DEVICES"] = ""

import numpy as np
import pytest
import tensorflow as tf

import zfit.z.math

prec = 0.00001


def test_polynomial():
    coeffs = [5.3, 1.2, complex(1.3, 0.4), -42, 32.4, 529.3, -0.93]
    x = tf.constant(5.)
    polynom_tf = zfit.z.math.poly_complex(*(coeffs + [x]))  # py34 comp: *x, y does not work
    polynom_np = np.polyval(coeffs[::-1], 5.)

    result = polynom_tf.numpy()
    assert result == pytest.approx(polynom_np, rel=prec)


def test_auto_upcast():
    tensor_from_f32 = _auto_upcast(tf.constant(5, dtype=tf.float32))
    tensor_from_f64 = _auto_upcast(tf.constant(5, dtype=tf.float64))
    assert tensor_from_f32.dtype == tf.float64
    assert tensor_from_f64.dtype == tf.float64

    tensor_from_i32 = _auto_upcast(tf.constant(5, dtype=tf.int32))
    tensor_from_i64 = _auto_upcast(tf.constant(5, dtype=tf.int64))
    assert tensor_from_i32.dtype == tf.int64
    assert tensor_from_i64.dtype == tf.int64

    tensor_from_c64 = _auto_upcast(tf.constant(5., dtype=tf.complex64))
    tensor_from_c128 = _auto_upcast(tf.constant(5., dtype=tf.complex128))
    assert tensor_from_c64.dtype == tf.complex128
    assert tensor_from_c128.dtype == tf.complex128



#  Copyright (c) 2021 zfit
import collections
import contextlib
import itertools
import math
import warnings
from collections import OrderedDict
from typing import (Callable, Dict, Iterable, List, Mapping, Optional, Tuple,
                    Union)

import colored
import iminuit
import numpy as np
import scipy.optimize
import scipy.stats
from colorama import Style, init
from ordered_set import OrderedSet
from scipy.optimize import LbfgsInvHessProduct
from tabulate import tabulate

from ..core.interfaces import ZfitIndependentParameter, ZfitLoss, ZfitParameter
from ..core.parameter import assign_values
from ..settings import run
from ..util.container import convert_to_container
from ..util.deprecation import deprecated_args
from ..util.warnings import ExperimentalFeatureWarning
from ..util.ztyping import ParamsTypeOpt
from .errors import (compute_errors, covariance_with_weights, dict_to_matrix,
                     matrix_to_dict)
from .interface import ZfitMinimizer, ZfitResult
from .termination import ConvergenceCriterion

init(autoreset=True)


class Approximations:

    def __init__(self, params: List[ZfitParameter], gradient: Optional[np.ndarray] = None,
                 hessian: Optional[np.ndarray] = None,
                 inv_hessian: Optional[np.ndarray] = None) -> None:
        """Holds different approximations after the minimisation and/or calculates them.

        Args:
            params: List of parameters the approximations (gradient, hessian, ...) were calculated with.
            gradient: Gradient
            hessian: Hessian Matrix
            inv_hessian: Inverse of the Hessian Matrix
        """
        self._params = params
        self._gradient = gradient
        self._hessian = hessian
        self._inv_hessian = inv_hessian
        super().__init__()

    @property
    def params(self):
        return self._params

    def gradient(self,
                 params: Optional[Union[ZfitParameter, Iterable[ZfitParameter]]] = None
                 ) -> Union[np.ndarray, None]:
        """Return an approximation of the gradient _if available_.

        Args:
            params: Parameters to which the gradients should be returned

        Returns:
            Array with gradients or `None`
        """
        grad = self._gradient
        if grad is None:
            return None

        if params is not None:
            params = convert_to_container(params, container=tuple)
            params_mapped = {i: params.index(param) for i, param in enumerate(self.params) if param in params}
            indices = sorted(params_mapped, key=lambda x: params_mapped[x])
            grad = grad[indices]
        return grad

    def hessian(self, invert: bool = True) -> Union[np.ndarray, None]:
        """Return an approximation of the hessian _if available_.

        Args:
            invert: If a _hessian approximation_ is not available but an inverse hessian is, invert the latter to
                obtain the hessian approximation.

        Returns:
            Array with hessian matrix or `None`
        """
        hess = self._hessian
        if hess is None and invert:
            inv_hess = self._inv_hessian
            if inv_hess is not None:
                hess = np.linalg.inv(inv_hess)
                self._hessian = hess
        return hess

    def inv_hessian(self, invert: bool = True) -> Union[None, np.ndarray]:
        """Return an approximation of the inverse hessian _if available_.

        Args:
            invert: If an _inverse hessian approximation_ is not available but a hessian is, invert the latter to
                obtain the inverse hessian approximation.

        Returns:
            Array with the inverse of the hessian matrix or `None`
        """
        inv_hess = self._inv_hessian
        if inv_hess is None and invert:
            hess = self._hessian
            if hess is not None:
                inv_hess = np.linalg.inv(hess)
                self._inv_hessian = inv_hess
        return inv_hess


def _minos_minuit(result, params, cl=None):
    minuit_minimizer = result._create_minuit_instance()

    try:
        minuit_minimizer.minos(*(p.name for p in params), cl=cl)
        # Minuit seems very ustable on this and the call can fail after a few trials
    except RuntimeError as error:
        if 'Function minimum is not valid.' not in error.args[0]:
            raise
        minuit_minimizer.reset()
        minuit_minimizer.minos(*(p.name for p in params), cl=cl)

    merror_result = minuit_minimizer.merrors  # returns every var
    attrs = ['lower', 'upper', 'is_valid', 'upper_valid', 'lower_valid', 'at_lower_limit', 'at_upper_limit', 'nfcn']
    errors = {}
    for p in params:
        error_res = merror_result[p.name]
        errors[p] = {attr: getattr(error_res, attr) for attr in attrs}
        errors[p]['original'] = error_res
    new_result = None
    return errors, new_result


def _covariance_minuit(result, params):
    minuit = result._create_minuit_instance()

    _ = minuit.hesse()  # make sure to have an accurate covariance

    covariance = minuit.covariance

    covariance_dict = {}
    if covariance is None:
        warnings.warn('minuit failed to calculate the covariance matrix or similar when calling `hesse`.'
                      'Try to use `hesse_np` as the method instead and try again.'
                      'This is unexpected and may has to do with iminuitV2. Either way, please fill an issue if'
                      ' this is not expected to fail for you.')
    else:
        for p1 in params:
            for p2 in params:
                key = (p1, p2)
                covariance_dict[key] = covariance[tuple(k.name for k in key)]

    return covariance_dict


def _covariance_np(result, params):
    if any(data.weights is not None for data in result.loss.data):
        warnings.warn("The computation of the covariance matrix with weights is still experimental.",
                      ExperimentalFeatureWarning)

    # TODO: maybe activate again? currently fails due to numerical problems
    # numgrad_was_none = settings.options.numerical_grad is None
    # if numgrad_was_none:
    #     settings.options.numerical_grad = True

    _, gradient, hessian = result.loss.value_gradient_hessian(params)
    covariance = np.linalg.inv(hessian)

    # if numgrad_was_none:
    #     settings.options.numerical_grad = None

    return matrix_to_dict(params, covariance)


def _covariance_approx(result, params):
    if any(data.weights is not None for data in result.loss.data):
        warnings.warn("Approximate covariance/hesse estimation with weights is not supported, returning None",
                      RuntimeWarning)
    covariance_dict = {}
    inv_hessian = result.approx.inv_hessian(invert=True)
    if inv_hessian is None:
        return covariance_dict

    params_approx = list(result.params)
    for p1 in params:
        p1_index = params_approx.index(p1)
        for p2 in params:
            p2_index = params_approx.index(p2)
            index = (p1_index, p2_index)
            key = (p1, p2)
            covariance_dict[key] = inv_hessian[index]
    return covariance_dict


class ParamToNameGetitem:
    __slots__ = ()

    def __getitem__(self, item):
        if isinstance(item, ZfitParameter):
            item = item.name
        return super().__getitem__(item)


class NameToParamGetitem:
    __slots__ = ()

    def __getitem__(self, item):
        if isinstance(item, str):
            for param in self.keys():
                if param.name == item:
                    item = param
                    break
        return super().__getitem__(item)


class FitResult(ZfitResult):
    _default_hesse = "hesse_np"
    _hesse_methods = {"minuit_hesse": _covariance_minuit, "hesse_np": _covariance_np, "approx": _covariance_approx}
    _default_error = "zfit_error"
    _error_methods = {"minuit_minos": _minos_minuit, "zfit_error": compute_errors}

    def __init__(self,
                 loss: ZfitLoss,
                 params: Dict[ZfitParameter, float],
                 minimizer: "ZfitMinimizer",
                 valid: bool,
                 edm: float,
                 fmin: float,
                 criterion: Optional[ConvergenceCriterion],
                 status: Optional[int] = None,
                 converged: Optional[bool] = None,
                 message: Optional[str] = None,
                 info: Optional[Mapping] = None,
                 approx: Optional[Union[Mapping, Approximations]] = None,
                 niter: Optional[int] = None,
                 evaluator: "zfit.minimizer.evaluation.LossEval" = None,
                 ) -> None:
        """Create a `FitResult` from a minimization. Store parameter values, minimization infos and calculate errors.

        Any errors calculated are saved under `self.params` dictionary with::

            {parameter: {error_name1: {'low': value, 'high': value or similar}}

        Args:
            loss: |@doc:result.init.loss| The loss function that was minimized.
               Usually, but not necessary, contains
               also the pdf, data and constraints. |@docend:result.init.loss|
            params: |@doc:result.init.params| Result of the fit where each
               :py:class:`~zfit.Parameter` key has the
               value from the minimum found by the minimizer. |@docend:result.init.params|
            minimizer: |@doc:result.init.minimizer| Minimizer that was used to obtain this `FitResult` and will be used to
                   calculate certain errors. If the minimizer
                   is state-based (like "iminuit"), then this is a copy
                   and the state of other `FitResults` or of the *actual*
                   minimizer that performed the minimization
                   won't be altered. |@docend:result.init.minimizer|
            valid: |@doc:result.init.valid| Indicating whether the result is valid or not. This is the strongest
                   indication and serves as
                   the global flag. The reasons why a result may be
                   invalid can be arbitrary, including but not exclusive:

                   - parameter(s) at the limit
                   - maxiter reached without proper convergence
                   - the minimizer may even converged but it is known that this is only a local minimum

                   To indicate the reason for the invalidity, pass a message. |@docend:result.init.valid|
            edm: |@doc:result.init.edm| The estimated distance to minimum
                   which is the criterion value at the minimum. |@docend:result.init.edm|
            fmin: |@doc:result.init.fmin| Value of the function at the minimum. |@docend:result.init.fmin|
            criterion: |@doc:result.init.criterion| Criterion that was used during the minimization.
                   This determines the estimated distance to the
                   minimum (edm) |@docend:result.init.criterion|
            status: |@doc:result.init.status| A status code (if available) that describes
                   the minimization termination. 0 means a valid
                   termination. |@docend:result.init.status|
            converged: |@doc:result.init.converged| Whether the fit has successfully converged or not.
                   The result itself can still be an invalid minimum
                   such as if the parameters are at or close
                   to the limits or in case another minimum is found. |@docend:result.init.converged|
            message: |@doc:result.init.message| Human readable message to indicate the reason
                   if the fitresult is not valid.
                   If the fit is valid, the message (should)
                   be an empty string (or None),
                   otherwise it should denote the reason for the invalidity. |@docend:result.init.message|
            info: |@doc:result.init.info| Additional information (if available)
                   such as *number of gradient function calls* or the
                   original minimizer return message.
                   This is a relatively free field and _no single field_
                   in it is guaranteed to be stable.
                   Some recommended fields:

                   - *original*: contains the original returned object
                     by the minimizer used internally.
                   - *optimizer*: the actual instance of the wrapped
                     optimizer (if available) |@docend:result.init.info|
            approx: |@doc:result.init.approx| Collection of approximations found during
                   the minimization process such as gradient and hessian. |@docend:result.init.approx|
            niter: |@doc:result.init.niter| Approximate number of iterations ~= number
                   of function evaluations ~= number of gradient evaluations.
                   This is an approximated value and the exact meaning
                   can differ between different minimizers. |@docend:result.init.niter|
            evaluator: |@doc:result.init.evaluator| Loss evaluator that was used during the
                   minimization and that may contains information
                   about the last evaluations of the gradient
                   etc which can serve as approximations. |@docend:result.init.evaluator|
        """
        super().__init__()

        if status is None:
            status = 0 if valid else -999
        if converged is None and valid:
            converged = True
        if message is None:
            if valid:
                message = ''
            else:
                message = "Invalid, unknown reason (not specified)"

        info = {} if info is None else info
        approx = self._input_convert_approx(approx, evaluator, info, params)

        if evaluator is not None:
            niter = evaluator.niter if niter is None else niter

        param_at_limit = any(param.at_limit for param in params)
        if param_at_limit:
            valid = False
            if message:
                message += ' AND '
            message += "parameter(s) at their limit."

        self._cache_minuit = None  # in case used in errors

        self._evaluator = evaluator  # keep private for now
        self._niter = niter  # keep private for now
        self._approx = approx
        self._status = status
        self._message = "" if message is None else message
        self._converged = converged
        self._params = self._input_convert_params(params)
        self._values = ValuesHolder(params)
        self._params_at_limit = param_at_limit
        self._edm = edm
        self._criterion = criterion
        self._fmin = fmin
        self._info = info
        self._loss = loss
        self._minimizer = minimizer
        self._valid = valid
        self._covariance_dict = {}

    def _input_convert_approx(self, approx, evaluator, info, params):
        """Convert approx (if a Mapping) to an `Approximation` using the information provided.

        Args:
            approx: |@doc:result.init.approx| Collection of approximations found during
                   the minimization process such as gradient and hessian. |@docend:result.init.approx|
            evaluator: |@doc:result.init.evaluator| Loss evaluator that was used during the
                   minimization and that may contains information
                   about the last evaluations of the gradient
                   etc which can serve as approximations. |@docend:result.init.evaluator|
            info: |@doc:result.init.info| Additional information (if available)
                   such as *number of gradient function calls* or the
                   original minimizer return message.
                   This is a relatively free field and _no single field_
                   in it is guaranteed to be stable.
                   Some recommended fields:

                   - *original*: contains the original returned object
                     by the minimizer used internally.
                   - *optimizer*: the actual instance of the wrapped
                     optimizer (if available) |@docend:result.init.info|
            params: |@doc:result.init.params| Result of the fit where each
               :py:class:`~zfit.Parameter` key has the
               value from the minimum found by the minimizer. |@docend:result.init.params|

        Returns:
            The created approximation.
        """
        approx = {} if approx is None else approx
        if isinstance(approx, collections.Mapping):
            if 'params' not in approx:
                approx['params'] = params

            if info:
                if 'gradient' not in approx:
                    approx['gradient'] = info.get('grad', info.get('gradient'))
                if 'hessian' not in approx:
                    approx['hessian'] = info.get('hess', info.get('hesse', info.get('hessian')))
                if 'inv_hessian' not in approx:
                    approx['inv_hessian'] = info.get('inv_hess', info.get('inv_hesse', info.get('inv_hessian')))
            if evaluator is not None:
                if 'gradient' not in approx:
                    approx['gradient'] = evaluator.last_gradient
                if 'hessian' not in approx:
                    approx['hessian'] = evaluator.last_hessian

            approx = Approximations(**approx)
        return approx

    def _input_convert_params(self, params):
        return ParamHolder((p, {"value": v}) for p, v in params.items())

    def _check_get_uncached_params(self, params, method_name, cl):
        uncached = []
        for p in params:
            errordict = self.params[p].get(method_name)
            # cl is < 1 and gets very close. The closer, the more it matters -> scale tolerance by it
            if errordict is not None and not math.isclose(errordict['cl'], cl, abs_tol=3e-3 * (1 - cl)):
                raise NameError(f"Error with name {method_name} already exists in {repr(self)} with a different"
                                f" convidence level of {errordict['cl']} instead of the requested {cl}.")
            else:
                uncached.append(p)
        return uncached

    def _create_minuit_instance(self):
        minuit = self._cache_minuit
        from zfit.minimizers.minimizer_minuit import Minuit
        if minuit is None:
            if isinstance(self.minimizer, Minuit):
                minuit = self.minimizer._minuit_minimizer
            else:
                minimizer = Minuit(tol=self.minimizer.tol, verbosity=0, name="ZFIT_TMP_UNCERTAINITIES")
                minuit, _, _ = minimizer._make_minuit(loss=self.loss, params=self.params, init=self)
            self._cache_minuit = minuit
        return minuit

    @classmethod
    def from_ipopt(cls,
                   loss: ZfitLoss,
                   params: Iterable[ZfitParameter],
                   problem: 'ipyopt.Problem',
                   minimizer: 'zfit.minimize.IpyoptV1',
                   valid: bool,
                   values: np.ndarray,
                   message: Optional[str],
                   converged: Optional[bool],
                   edm: Union['zfit.minimizers.termination.CriterionNotAvailable', float],
                   niter: Optional[int],
                   fmin: Optional[float],
                   status: Optional[int],
                   criterion: 'zfit.minimizers.termination.ConvergenceCriterion',
                   evaluator: Optional['zfit.minimizers.evaluation.LossEval'],
                   ) -> 'FitResult':
        """Create a ``FitResult`` from an ipopt minimization.

        Args:
            loss: |@doc:result.init.loss| The loss function that was minimized.
               Usually, but not necessary, contains
               also the pdf, data and constraints. |@docend:result.init.loss|
            params: |@doc:result.init.params| Result of the fit where each
               :py:class:`~zfit.Parameter` key has the
               value from the minimum found by the minimizer. |@docend:result.init.params|
            problem: |@doc:result.init.problem||@docend:result.init.problem|
            minimizer: |@doc:result.init.minimizer| Minimizer that was used to obtain this `FitResult` and will be used to
                   calculate certain errors. If the minimizer
                   is state-based (like "iminuit"), then this is a copy
                   and the state of other `FitResults` or of the *actual*
                   minimizer that performed the minimization
                   won't be altered. |@docend:result.init.minimizer|
            valid: |@doc:result.init.valid| Indicating whether the result is valid or not. This is the strongest
                   indication and serves as
                   the global flag. The reasons why a result may be
                   invalid can be arbitrary, including but not exclusive:

                   - parameter(s) at the limit
                   - maxiter reached without proper convergence
                   - the minimizer may even converged but it is known that this is only a local minimum

                   To indicate the reason for the invalidity, pass a message. |@docend:result.init.valid|
            values: |@doc:result.init.values| Values of the parameters at the
                   found minimum. |@docend:result.init.values|
            message: |@doc:result.init.message| Human readable message to indicate the reason
                   if the fitresult is not valid.
                   If the fit is valid, the message (should)
                   be an empty string (or None),
                   otherwise it should denote the reason for the invalidity. |@docend:result.init.message|
            converged: |@doc:result.init.converged| Whether the fit has successfully converged or not.
                   The result itself can still be an invalid minimum
                   such as if the parameters are at or close
                   to the limits or in case another minimum is found. |@docend:result.init.converged|
            edm: |@doc:result.init.edm| The estimated distance to minimum
                   which is the criterion value at the minimum. |@docend:result.init.edm|
            niter: |@doc:result.init.niter| Approximate number of iterations ~= number
                   of function evaluations ~= number of gradient evaluations.
                   This is an approximated value and the exact meaning
                   can differ between different minimizers. |@docend:result.init.niter|
            fmin: |@doc:result.init.fmin| Value of the function at the minimum. |@docend:result.init.fmin|
            status: |@doc:result.init.status| A status code (if available) that describes
                   the minimization termination. 0 means a valid
                   termination. |@docend:result.init.status|
            criterion: |@doc:result.init.criterion| Criterion that was used during the minimization.
                   This determines the estimated distance to the
                   minimum (edm) |@docend:result.init.criterion|
            evaluator: |@doc:result.init.evaluator| Loss evaluator that was used during the
                   minimization and that may contains information
                   about the last evaluations of the gradient
                   etc which can serve as approximations. |@docend:result.init.evaluator|

        Returns:
            ``zfit.minimize.FitResult``:
        """
        info = {'problem': problem}
        params = dict(zip(params, values))
        valid = valid if converged is None else valid and converged
        if evaluator is not None:
            valid = valid and not evaluator.maxiter_reached
        return cls(params=params, loss=loss, fmin=fmin, edm=edm, message=message,
                   criterion=criterion, info=info, valid=valid, converged=converged,
                   niter=niter, status=status, minimizer=minimizer, evaluator=evaluator)

    @classmethod
    def from_minuit(cls,
                    loss: ZfitLoss,
                    params: Iterable[ZfitParameter],
                    minuit: iminuit.Minuit,
                    minimizer: Union[ZfitMinimizer, iminuit.Minuit],
                    valid: Optional[bool],
                    values: Optional[np.ndarray] = None,
                    message: Optional[str] = None,
                    converged: Optional[bool] = None,
                    edm: Optional[Union['zfit.minimizers.termination.CriterionNotAvailable', float]] = None,
                    niter: Optional[int] = None,
                    fmin: Optional[float] = None,
                    status: Optional[int] = None,
                    criterion: Optional['zfit.minimizers.termination.ConvergenceCriterion'] = None,
                    evaluator: Optional['zfit.minimizers.evaluation.LossEval'] = None) -> 'FitResult':
        """Create a `FitResult` from a :py:class:`~iminuit.util.MigradResult` returned by
        :py:meth:`iminuit.Minuit.migrad` and a iminuit :py:class:`~iminuit.Minuit` instance with the corresponding
        zfit objects.

        Args:
            loss: zfit Loss that was minimized.
            params: Iterable of the zfit parameters that were floating during the minimization.
            minuit: Return value of the iminuit migrad command, the instance of :class:`iminuit.Minuit`
            minimizer: Instance of the zfit Minuit minimizer that was used to minimize the loss.
            valid: |@doc:result.init.valid| Indicating whether the result is valid or not. This is the strongest
                   indication and serves as
                   the global flag. The reasons why a result may be
                   invalid can be arbitrary, including but not exclusive:

                   - parameter(s) at the limit
                   - maxiter reached without proper convergence
                   - the minimizer may even converged but it is known that this is only a local minimum

                   To indicate the reason for the invalidity, pass a message. |@docend:result.init.valid|
            values: |@doc:result.init.values| Values of the parameters at the
                   found minimum. |@docend:result.init.values|
            message: |@doc:result.init.message| Human readable message to indicate the reason
                   if the fitresult is not valid.
                   If the fit is valid, the message (should)
                   be an empty string (or None),
                   otherwise it should denote the reason for the invalidity. |@docend:result.init.message|
            converged: |@doc:result.init.converged| Whether the fit has successfully converged or not.
                   The result itself can still be an invalid minimum
                   such as if the parameters are at or close
                   to the limits or in case another minimum is found. |@docend:result.init.converged|
            edm: |@doc:result.init.edm| The estimated distance to minimum
                   which is the criterion value at the minimum. |@docend:result.init.edm|
            niter: |@doc:result.init.niter| Approximate number of iterations ~= number
                   of function evaluations ~= number of gradient evaluations.
                   This is an approximated value and the exact meaning
                   can differ between different minimizers. |@docend:result.init.niter|
            fmin: |@doc:result.init.fmin| Value of the function at the minimum. |@docend:result.init.fmin|
            status: |@doc:result.init.status| A status code (if available) that describes
                   the minimization termination. 0 means a valid
                   termination. |@docend:result.init.status|
            criterion: |@doc:result.init.criterion| Criterion that was used during the minimization.
                   This determines the estimated distance to the
                   minimum (edm) |@docend:result.init.criterion|
            evaluator: |@doc:result.init.evaluator| Loss evaluator that was used during the
                   minimization and that may contains information
                   about the last evaluations of the gradient
                   etc which can serve as approximations. |@docend:result.init.evaluator|


        Returns:
            ``zfit.minimize.FitResult``: A `FitResult` as if zfit Minuit was used.
        """
        from .minimizer_minuit import Minuit
        from .termination import EDM

        if not isinstance(minimizer, Minuit):
            if isinstance(minimizer, iminuit.Minuit):
                minimizer_new = Minuit()
                minimizer_new._minuit_minimizer = minimizer
                minimizer = minimizer_new
            else:
                raise ValueError(f"Minimizer {minimizer} not supported. Use `Minuit` from zfit or from iminuit.")

        params_result = [p_dict for p_dict in minuit.params]

        fmin_object = minuit.fmin
        minuit_converged = not fmin_object.is_above_max_edm
        converged = minuit_converged if converged is None else (converged and minuit_converged)
        niter = fmin_object.nfcn if niter is None else niter
        info = {'n_eval': niter,
                # 'grad': result['jac'],
                # 'message': result['message'],
                'minuit': minuit,
                'original': fmin_object}
        if fmin_object.has_covariance:
            info['inv_hessian'] = np.array(minuit.covariance)

        edm = fmin_object.edm if edm is None else edm
        if criterion is None:
            criterion = EDM(tol=minimizer.tol, loss=loss, params=params)
            criterion.last_value = edm
        fmin = fmin_object.fval if fmin is None else fmin
        minuit_valid = fmin_object.is_valid
        valid = minuit_valid if valid is None else minuit_valid and valid
        if evaluator is not None:
            valid = valid and not evaluator.maxiter_reached
        if values is None:
            values = (res.value for res in params_result)
        params = dict(zip(params, values))
        return cls(params=params, edm=edm, fmin=fmin, info=info, loss=loss, niter=niter, converged=converged,
                   status=status, message=message, valid=valid, criterion=criterion,
                   minimizer=minimizer, evaluator=evaluator)

    @classmethod
    def from_scipy(cls,
                   loss: ZfitLoss,
                   params: Iterable[ZfitParameter],
                   result: scipy.optimize.OptimizeResult,
                   minimizer: ZfitMinimizer,
                   message: Optional[str],
                   valid: bool,
                   criterion: ConvergenceCriterion,
                   edm: Optional[float] = None,
                   niter: Optional[int] = None,
                   evaluator: Optional['zfit.minimize.LossEval'] = None) -> 'FitResult':
        """Create a ``FitResult from a SciPy `~scipy.optimize.OptimizeResult`.

        Args:
            loss: |@doc:result.init.loss| The loss function that was minimized.
               Usually, but not necessary, contains
               also the pdf, data and constraints. |@docend:result.init.loss|
            params: |@doc:result.init.params| Result of the fit where each
               :py:class:`~zfit.Parameter` key has the
               value from the minimum found by the minimizer. |@docend:result.init.params|
            result: Result of the SciPy optimization.
            minimizer: |@doc:result.init.minimizer| Minimizer that was used to obtain this `FitResult` and will be used to
                   calculate certain errors. If the minimizer
                   is state-based (like "iminuit"), then this is a copy
                   and the state of other `FitResults` or of the *actual*
                   minimizer that performed the minimization
                   won't be altered. |@docend:result.init.minimizer|
            message: |@doc:result.init.message| Human readable message to indicate the reason
                   if the fitresult is not valid.
                   If the fit is valid, the message (should)
                   be an empty string (or None),
                   otherwise it should denote the reason for the invalidity. |@docend:result.init.message|
            edm: |@doc:result.init.edm| The estimated distance to minimum
                   which is the criterion value at the minimum. |@docend:result.init.edm|
            niter: |@doc:result.init.niter| Approximate number of iterations ~= number
                   of function evaluations ~= number of gradient evaluations.
                   This is an approximated value and the exact meaning
                   can differ between different minimizers. |@docend:result.init.niter|
            valid: |@doc:result.init.valid| Indicating whether the result is valid or not. This is the strongest
                   indication and serves as
                   the global flag. The reasons why a result may be
                   invalid can be arbitrary, including but not exclusive:

                   - parameter(s) at the limit
                   - maxiter reached without proper convergence
                   - the minimizer may even converged but it is known that this is only a local minimum

                   To indicate the reason for the invalidity, pass a message. |@docend:result.init.valid|
            criterion: |@doc:result.init.criterion| Criterion that was used during the minimization.
                   This determines the estimated distance to the
                   minimum (edm) |@docend:result.init.criterion|
            evaluator: |@doc:result.init.evaluator| Loss evaluator that was used during the
                   minimization and that may contains information
                   about the last evaluations of the gradient
                   etc which can serve as approximations. |@docend:result.init.evaluator|

        Returns:
            `zfit.minimize.FitResult`:
        """
        result_values = result['x']
        if niter is None:
            niter = result.get('nit')

        converged = result.get('success', valid)
        status = result['status']
        inv_hesse = result.get('hess_inv')
        if isinstance(inv_hesse, LbfgsInvHessProduct):
            inv_hesse = inv_hesse.todense()
        if message is None and (not converged or not valid):
            message = result.get('message')
        grad = result.get('grad')
        info = {'n_eval': result['nfev'],
                'n_iter': niter,
                'niter': niter,
                'grad': result.get('jac') if grad is None else grad,
                'inv_hesse': inv_hesse,
                'hesse': result.get('hesse'),
                'message': message,
                'evaluator': evaluator,
                'original': result}
        approx = dict(params=params,
                      gradient=info.get('grad'),
                      hessian=info.get('hesse'),
                      inv_hessian=info.get('inv_hesse'))

        fmin = result['fun']
        params = dict(zip(params, result_values))
        if evaluator is not None:
            valid = valid and not evaluator.maxiter_reached

        fitresult = cls(params=params, edm=edm, fmin=fmin, info=info, approx=approx,
                        converged=converged, status=status, message=message, valid=valid, niter=niter,
                        loss=loss, minimizer=minimizer, criterion=criterion, evaluator=evaluator)
        return fitresult

    @classmethod
    def from_nlopt(cls,
                   loss: ZfitLoss,
                   opt,
                   params: Iterable[ZfitParameter],
                   minimizer: Union[ZfitMinimizer, iminuit.Minuit],
                   valid: Optional[bool],
                   values: Optional[np.ndarray] = None,
                   message: Optional[str] = None,
                   converged: Optional[bool] = None,
                   edm: Optional[Union['zfit.minimizers.termination.CriterionNotAvailable', float]] = None,
                   niter: Optional[int] = None,
                   fmin: Optional[float] = None,
                   status: Optional[int] = None,
                   criterion: Optional['zfit.minimizers.termination.ConvergenceCriterion'] = None,
                   evaluator: Optional['zfit.minimizers.evaluation.LossEval'] = None,
                   inv_hessian: Optional[np.ndarray] = None,
                   hessian: Optional[np.ndarray] = None,
                   ) -> 'FitResult':
        """Create a ``FitResult`` from an NLopt optimizer.

        Args:
            loss: |@doc:result.init.loss| The loss function that was minimized.
               Usually, but not necessary, contains
               also the pdf, data and constraints. |@docend:result.init.loss|
            opt: Optimizer instance of NLopt
            params: |@doc:result.init.params| Result of the fit where each
               :py:class:`~zfit.Parameter` key has the
               value from the minimum found by the minimizer. |@docend:result.init.params|
            minimizer: |@doc:result.init.minimizer| Minimizer that was used to obtain this `FitResult` and will be used to
                   calculate certain errors. If the minimizer
                   is state-based (like "iminuit"), then this is a copy
                   and the state of other `FitResults` or of the *actual*
                   minimizer that performed the minimization
                   won't be altered. |@docend:result.init.minimizer|
            valid: |@doc:result.init.valid| Indicating whether the result is valid or not. This is the strongest
                   indication and serves as
                   the global flag. The reasons why a result may be
                   invalid can be arbitrary, including but not exclusive:

                   - parameter(s) at the limit
                   - maxiter reached without proper convergence
                   - the minimizer may even converged but it is known that this is only a local minimum

                   To indicate the reason for the invalidity, pass a message. |@docend:result.init.valid|
            values: |@doc:result.init.values| Values of the parameters at the
                   found minimum. |@docend:result.init.values|
            message: |@doc:result.init.message| Human readable message to indicate the reason
                   if the fitresult is not valid.
                   If the fit is valid, the message (should)
                   be an empty string (or None),
                   otherwise it should denote the reason for the invalidity. |@docend:result.init.message|
            converged: |@doc:result.init.converged| Whether the fit has successfully converged or not.
                   The result itself can still be an invalid minimum
                   such as if the parameters are at or close
                   to the limits or in case another minimum is found. |@docend:result.init.converged|
            edm: |@doc:result.init.edm| The estimated distance to minimum
                   which is the criterion value at the minimum. |@docend:result.init.edm|
            niter: |@doc:result.init.niter| Approximate number of iterations ~= number
                   of function evaluations ~= number of gradient evaluations.
                   This is an approximated value and the exact meaning
                   can differ between different minimizers. |@docend:result.init.niter|
            fmin: |@doc:result.init.fmin| Value of the function at the minimum. |@docend:result.init.fmin|
            status: |@doc:result.init.status| A status code (if available) that describes
                   the minimization termination. 0 means a valid
                   termination. |@docend:result.init.status|
            criterion: |@doc:result.init.criterion| Criterion that was used during the minimization.
                   This determines the estimated distance to the
                   minimum (edm) |@docend:result.init.criterion|
            evaluator: |@doc:result.init.evaluator| Loss evaluator that was used during the
                   minimization and that may contains information
                   about the last evaluations of the gradient
                   etc which can serve as approximations. |@docend:result.init.evaluator|
            inv_hessian: The (approximated) inverse hessian matrix.
            hessian: The (approximated) hessian matrix.

        Returns:
            zfit.minimizers.fitresult.FitResult:
        """
        converged = converged if converged is None else bool(converged)
        param_dict = {p: v for p, v in zip(params, values)}
        if fmin is None:
            fmin = opt.last_optimum_value()
        status_nlopt = opt.last_optimize_result()
        if status is None:
            status = status_nlopt
        niter = opt.get_numevals() if niter is None else niter
        converged = 1 <= status_nlopt <= 4 and converged is not False

        messages = {
            1: "NLOPT_SUCCESS",
            2: "NLOPT_STOPVAL_REACHED",
            3: "NLOPT_FTOL_REACHED",
            4: "NLOPT_XTOL_REACHED",
            5: "NLOPT_MAXEVAL_REACHED",
            6: "NLOPT_MAXTIME_REACHED",
            -1: "NLOPT_FAILURE",
            -2: "NLOPT_INVALID_ARGS",
            -3: "NLOPT_OUT_OF_MEMORY",
            -4: "NLOPT_ROUNDOFF_LIMITED",
            -5: "NLOPT_FORCED_STOP",
        }
        message_nlopt = messages[status_nlopt]
        info = {'n_eval': niter,
                'niter': niter,
                'message': message_nlopt,
                'original': status,
                'evaluator': evaluator,
                'status': status}
        if message is None:
            message = message_nlopt

        valid = valid and converged
        if evaluator is not None:
            valid = valid and not evaluator.maxiter_reached

        approx = {}
        if inv_hessian is None:
            if hessian is None and evaluator is not None:
                hessian = evaluator.last_hessian
            if hessian is not None:
                inv_hessian = np.linalg.inv(hessian)

        if inv_hessian is not None:
            info['inv_hesse'] = inv_hessian
            approx['inv_hessian'] = inv_hessian

        return cls(params=param_dict,
                   edm=edm,
                   fmin=fmin,
                   status=status,
                   converged=converged,
                   info=info,
                   niter=niter,
                   valid=valid,
                   loss=loss,
                   minimizer=minimizer,
                   criterion=criterion,
                   message=message,
                   evaluator=evaluator)

    @property
    def approx(self) -> Approximations:
        return self._approx

    @property
    def params(self) -> Mapping[ZfitIndependentParameter, Mapping[str, Mapping[str, object]]]:
        return self._params

    @property
    def values(self) -> Mapping[Union[str, ZfitParameter], float]:
        return self._values

    @property
    def criterion(self) -> ConvergenceCriterion:
        return self._criterion

    @property
    def message(self) -> str:
        return self._message

    @property
    def edm(self) -> float:
        """The estimated distance to the minimum.

        Returns:
            Numeric
        """
        return self._edm

    @property
    def minimizer(self) -> ZfitMinimizer:
        return self._minimizer

    @property
    def loss(self) -> ZfitLoss:
        # TODO(Mayou36): this is currently a reference, should be a copy of the loss?
        return self._loss

    @property
    def fmin(self) -> float:
        """Function value at the minimum.

        Returns:
            Numeric
        """
        return self._fmin

    @property
    def status(self):
        return self._status

    @property
    def info(self) -> Mapping[str, object]:
        return self._info

    @property
    def converged(self) -> bool:
        return self._converged

    @property
    def valid(self) -> bool:
        return self._valid and not self.params_at_limit and self.converged

    @property
    def params_at_limit(self) -> bool:
        return self._params_at_limit

    @contextlib.contextmanager
    def _input_check_reset_params(self, params):
        params = self._input_check_params(params=params)
        old_values = run(params)
        try:
            yield params
        except Exception:
            warnings.warn("Exception occurred, parameter values are not reset and in an arbitrary, last"
                          " used state. If this happens during normal operation, make sure you reset the values.",
                          RuntimeWarning)
            raise
        assign_values(params=params, values=old_values)  # TODO: or set?

    def _input_check_params(self, params):
        if params is not None:
            params = convert_to_container(params)
        else:
            params = list(self.params.keys())
        return params

    @deprecated_args(None, "Use `name` instead", "error_name")
    def hesse(self,
              params: ParamsTypeOpt = None,
              method: Union[str, Callable] = None,
              cl: Optional[float] = None,
              name: Optional[Union[str, bool]] = None,
              error_name: Optional[str] = None
              ) -> Dict[ZfitIndependentParameter, Dict]:
        """Calculate for `params` the symmetric error using the Hessian/covariance matrix.

        Args:
            cl:
            params: The parameters to calculate the
                Hessian symmetric error. If None, use all parameters.
            method: the method to calculate the covariance matrix. Can be
                {'minuit_hesse', 'hesse_np', 'approx'} or a callable.
            name: The name for the error in the dictionary. This will be added to
                the information collected in params under ``params[p][name]`` where
                p is a Parameter. If the name is `False`, it won't be added and only
                returned.
            error_name: The name for the error in the dictionary.

        Returns:
            Result of the hessian (symmetric) error as dict with each parameter holding
                the error dict {'error': sym_error}.

                So given param_a (from zfit.Parameter(.))
                `error_a = result.hesse(params=param_a)[param_a]['error']`
                error_a is the hessian error.
        """
        # for compatibility with `errors`
        cl = 0.68268949 if cl is None else cl  # scipy.stats.chi2(1).cdf(1)
        if cl >= 1:
            raise ValueError(f"cl is the confidence limit and has to be < 1, not {cl}")

        if method is None:
            # LEGACY START
            method = self._default_hesse
            from zfit.minimizers.minimizer_minuit import Minuit

            if isinstance(self.minimizer, Minuit):
                method = "minuit_hesse"
            # LEGACY END
        # Deprecated name
        if error_name is not None:
            name = error_name
        if name is None:
            if not isinstance(method, str):
                raise ValueError("Need to specify `name` or use a string as `method`")
            name = method

        with self._input_check_reset_params(params) as params:
            uncached_params = self._check_get_uncached_params(params=params, method_name=name, cl=cl)
            if uncached_params:
                error_dict = self._hesse(params=uncached_params, method=method, cl=cl)
                if any(val['error'] is None for val in error_dict.values()):
                    return {}
                for p in error_dict:
                    error_dict[p]['cl'] = cl
                if name:
                    self._cache_errors(error_name=name, errors=error_dict)
            else:
                error_dict = {}

        error_dict.update({p: self.params[p][name] for p in params if p not in uncached_params})
        return {p: error_dict[p] for p in params}

    def _cache_errors(self, error_name, errors):
        for param, error in errors.items():
            self.params[param][error_name] = error

    def _hesse(self, params, method, cl):
        pseudo_sigma = scipy.stats.chi2(1).ppf(cl) ** 0.5

        covariance_dict = self.covariance(params, method, as_dict=True)
        return {
            p: {"error": covariance_dict[(p, p)] ** 0.5 * pseudo_sigma if covariance_dict[(p, p)] is not None else None}
            for p in params}

    def error(self, params: ParamsTypeOpt = None, method: Union[str, Callable] = None, error_name: str = None,
              sigma: float = 1.0) -> OrderedDict:
        r"""

        .. deprecated:: unknown
            Use :func:`errors` instead.

        Args:
            params: The parameters or their names to calculate the
                 errors. If `params` is `None`, use all *floating* parameters.
            method: The method to use to calculate the errors. Valid choices are
                {'minuit_minos'} or a Callable.
            sigma: Errors are calculated with respect to `sigma` std deviations. The definition
                of 1 sigma depends on the loss function and is defined there.

                For example, the negative log-likelihood (without the factor of 2) has a correspondents
                of :math:`\Delta` NLL of 1 corresponds to 1 std deviation.
            error_name: The name for the error in the dictionary.


        Returns:
            A `OrderedDict` containing as keys the parameter names and as value a `dict` which
                contains (next to probably more things) two keys 'lower' and 'upper',
                holding the calculated errors.
                Example: result['par1']['upper'] -> the asymmetric upper error of 'par1'
        """
        warnings.warn("`error` is deprecated, use `errors` instead. This will return not only the errors but also "
                      "(a possible) new FitResult if a minimum was found. So change"
                      "errors = result.error()"
                      "to"
                      "errors, new_res = result.errors()", DeprecationWarning)
        return self.errors(params=params, method=method, name=error_name, sigma=sigma)[0]

    @deprecated_args(None, "Use name instead.", 'error_name')
    @deprecated_args(None, "Use cl for confidence level instead.", 'sigma')
    def errors(self, params: ParamsTypeOpt = None, method: Union[str, Callable] = None, name: str = None,
               error_name: str = None,
               cl: Optional[float] = None, sigma=None) -> Tuple[OrderedDict, Union[None, 'FitResult']]:
        r"""Calculate and set for `params` the asymmetric error using the set error method.

            Args:
                params: The parameters or their names to calculate the
                     errors. If `params` is `None`, use all *floating* parameters.
                method: The method to use to calculate the errors. Valid choices are
                    {'minuit_minos'} or a Callable.
                cl: Uncertainties are calculated with respect to the confidence level cl. The default is 68.3%.
                    For example, the negative log-likelihood (without the factor of 2) has a correspondents
                    of :math:`\Delta` NLL of 1 corresponds to 1 std deviation.
                sigma: Errors are calculated with respect to `sigma` std deviations. The definition
                    of 1 sigma depends on the loss function and is defined there.
                name: The name for the error in the dictionary.
                error_name: The name for the error in the dictionary.


            Returns:
                A `OrderedDict` containing as keys the parameter and as value a `dict` which
                    contains (next to often more things) two keys 'lower' and 'upper',
                    holding the calculated errors. Furthermore it has `cl` to indicate the convidence level
                    the uncertainty was calculated with.
                    Example: result[par1]['upper'] -> the asymmetric upper error of 'par1'

        """
        # Deprecated name
        if error_name is not None:
            name = error_name

        if sigma is not None:
            if cl is not None:
                raise ValueError("Cannot define sigma and cl, use cl only.")
            else:
                cl = scipy.stats.chi2(1).cdf(sigma)

        if cl is None:
            cl = 0.68268949  # scipy.stats.chi2(1).cdf(1)

        if method is None:
            # TODO: legacy, remove 0.6
            from zfit.minimize import Minuit
            if isinstance(self.minimizer, Minuit):
                method = 'minuit_minos'
                warnings.warn("'minuit_minos' will be changed as the default errors method to a custom implementation"
                              "with the same functionality. If you want to make sure that 'minuit_minos' will be used "
                              "in the future, add it explicitly as in `errors(method='minuit_minos')`", FutureWarning)
            else:
                method = self._default_error
        if name is None:
            if not isinstance(method, str):
                raise ValueError("Need to specify `error_name` or use a string as `method`")
            name = method

        if method == 'zfit_error':
            warnings.warn(
                "'zfit_error' is still rather new. If it fails, please report it here:"
                " https://github.com/zfit/zfit/issues/new?assignees=&labels=bug&template"
                "=bug_report.md&title=zfit%20error%20fails.",
                ExperimentalFeatureWarning, stacklevel=2)

        params = self._input_check_params(params)

        with self._input_check_reset_params(self.params.keys()):
            # TODO: cache with cl!
            uncached_params = self._check_get_uncached_params(params=params, method_name=name, cl=cl)

            new_result = None

            if uncached_params:
                error_dict, new_result = self._error(params=uncached_params, method=method, cl=cl)
                for p in error_dict:
                    error_dict[p]['cl'] = cl
                self._cache_errors(error_name=name, errors=error_dict)

                if new_result is not None:
                    msg = "Invalid, a new minimum was found."
                    self._cache_errors(error_name=name, errors={p: msg for p in params})
                    self._valid = False
                    self._message = msg
                    new_result._cache_errors(error_name=name, errors=error_dict)
        all_errors = OrderedDict((p, self.params[p][name]) for p in params)

        return all_errors, new_result

    def _error(self, params, method, cl):
        if not callable(method):
            try:
                method = self._error_methods[method]
            except KeyError:
                raise KeyError(f"The following method is not a valid, implemented method: {method}")
        return method(result=self, params=params, cl=cl)

    def covariance(self, params: ParamsTypeOpt = None, method: Union[str, Callable] = None, as_dict: bool = False):
        """Calculate the covariance matrix for `params`.

        Args:
            params: The parameters to calculate
                the covariance matrix. If `params` is `None`, use all *floating* parameters.
            method: The method to use to calculate the covariance matrix. Valid choices are
                {'minuit_hesse', 'hesse_np'} or a Callable.
            as_dict: Default `False`. If `True` then returns a dictionnary.

        Returns:
            2D `numpy.array` of shape (N, N);
            `dict`(param1, param2) -> covariance if `as_dict == True`.
        """
        if method is None:
            # LEGACY START
            method = self._default_hesse
            from zfit.minimizers.minimizer_minuit import Minuit

            if isinstance(self.minimizer, Minuit):
                method = "minuit_hesse"
            # LEGACY END

        if method not in self._covariance_dict:
            with self._input_check_reset_params(params) as params:
                self._covariance_dict[method] = self._covariance(method=method)

        params = self._input_check_params(params)
        covariance = {k: self._covariance_dict[method].get(k) for k in itertools.product(params, params)}

        if as_dict:
            return covariance
        else:
            return dict_to_matrix(params, covariance)

    def _covariance(self, method):
        if not callable(method):
            try:
                method = self._hesse_methods[method]
            except KeyError:
                raise KeyError(f"The following method is not a valid, implemented method: {method}")

        params = list(self.params.keys())

        if any(data.weights is not None for data in self.loss.data):
            return covariance_with_weights(method=method, result=self, params=params)
        else:
            return method(result=self, params=params)

    def correlation(self, params: ParamsTypeOpt = None, method: Union[str, Callable] = None, as_dict: bool = False):
        """Calculate the correlation matrix for `params`.

        Args:
            params: The parameters to calculate
                the correlation matrix. If `params` is `None`, use all *floating* parameters.
            method: The method to use to calculate the correlation matrix. Valid choices are
                {'minuit_hesse', 'hesse_np'} or a Callable.
            as_dict: Default `False`. If `True` then returns a dictionnary.

        Returns:
            2D `numpy.array` of shape (N, N);
            `dict`(param1, param2) -> correlation if `as_dict == True`.
        """

        covariance = self.covariance(params=params, method=method, as_dict=False)
        correlation = covariance_to_correlation(covariance)

        if as_dict:
            params = self._input_check_params(params)
            return matrix_to_dict(params, correlation)
        else:
            return correlation

    def __str__(self):
        string = Style.BRIGHT + f'FitResult' + Style.NORMAL + f' of\n{self.loss} \nwith\n{self.minimizer}\n\n'
        string += tabulate(
            [[color_on_bool(self.valid), color_on_bool(self.converged, on_true=False),
              color_on_bool(self.params_at_limit, on_true=colored.bg(9), on_false=False),
              format_value(self.edm, highprec=False),
              format_value(self.fmin)]],
            ['valid', 'converged', 'param at limit', 'edm', 'min value'],
            tablefmt='fancy_grid',
            disable_numparse=True)
        string += '\n\n' + Style.BRIGHT + "Parameters\n" + Style.NORMAL
        string += str(self.params)
        return string

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(self.__repr__())
            return
        p.text(self.__str__())


def covariance_to_correlation(covariance):
    diag = np.diag(1 / np.diag(covariance) ** 0.5)
    return np.matmul(diag, np.matmul(covariance, diag))


def format_value(value, highprec=True):
    try:
        import iminuit
        m_error_class = iminuit.util.MError
    except ImportError:
        m_error_class = dict

    if isinstance(value, dict) and 'error' in value:
        value = value['error']
        value = f"{value:> 6.2g}"
        value = f'+/-{" " * (8 - len(value))}' + value
    if isinstance(value, m_error_class) or (isinstance(value, dict) and 'lower' in value and 'upper' in value):
        if isinstance(value, m_error_class):
            lower = value.lower
            upper = value.upper
        else:
            lower = value['lower']
            upper = value['upper']
        lower_sign = f"{np.sign(lower): >+}"[0]
        upper_sign = f"{np.sign(upper): >+}"[0]
        lower, upper = f"{np.abs(lower): >6.2g}", f"{upper: >6.2g}"
        lower = lower_sign + " " * (7 - len(lower)) + lower
        upper = upper_sign + " " * (7 - len(upper)) + upper
        # lower += " t" * (11 - len(lower))
        value = lower + " " * 3 + upper

    if isinstance(value, float):
        if highprec:
            value = f"{value:> 6.4g}"
        else:
            value = f"{value:> 6.2g}"
    return value


def color_on_bool(value, on_true=colored.bg(10), on_false=colored.bg(9)):
    if not value and on_false:
        value_add = on_false
    elif value and on_true:
        value_add = on_true
    else:
        value_add = ''
    value = value_add + str(value) + Style.RESET_ALL
    return value


class ListWithKeys(collections.UserList):
    __slots__ = ('_initdict',)

    def __init__(self, initdict) -> None:
        super().__init__(initlist=initdict.values())
        self._initdict = initdict

    def __getitem__(self, item):
        if isinstance(item, ZfitParameter):
            return self._initdict[item]
        return super().__getitem__(item)

    def keys(self):
        return self._initdict.keys()


class ValuesHolder(NameToParamGetitem, ListWithKeys):
    __slots__ = ()


class ParamHolder(NameToParamGetitem, collections.UserDict):

    def __str__(self) -> str:
        order_keys = ['value', 'hesse']
        keys = OrderedSet()
        for pdict in self.values():
            keys.update(OrderedSet(pdict))
        order_keys = OrderedSet([key for key in order_keys if key in keys])
        order_keys.update(keys)

        rows = []
        for param, pdict in self.items():
            row = [param.name]
            row.extend(format_value(pdict.get(key, ' ')) for key in order_keys)
            row.append(color_on_bool(run(param.at_limit), on_true=colored.bg('light_red'), on_false=False))
            rows.append(row)

        order_keys = ['name'] + list(order_keys) + ['at limit']
        table = tabulate(rows, order_keys, numalign="right", stralign='right', colalign=('left',))
        return table
